from mxcubecore.HardwareObjects.abstract.sample_changer.CatsMaint import CatsMaint


class ALBACatsMaint(CatsMaint):
    def __init__(self, *args):
        CatsMaint.__init__(self, *args)

    def init(self):
        CatsMaint.init(self)

        # load ALBA attributes and commands from XML
        self._chnAtHome = self.get_channel_object("_chnAtHome")
        self.super_abort_cmd = self.get_command_object("super_abort")

        # channel to ask diffractometer for mounting position
        self.shifts_channel = self.get_channel_object("shifts")

    def _do_abort(self):
        if self.super_abort_cmd is not None:
            self.super_abort_cmd()
        self._cmdAbort()

    def _do_reset_memory(self):
        """
        Reset CATS memory.
        """
        # Check do_PRO6_RAH first
        if self._chnAtHome.get_value() is True:
            CatsMaint._do_reset_memory(self)

    def _do_reset(self):
        """
        Reset CATS system.
        """
        self._cmdAbort()
        self._cmdReset()
        self._do_reset_memory()

    def _do_operation_command(self, cmd, pars):
        """
        Send a CATS command

        @cmd: command
        @pars: command arguments
        """
        CatsMaint._do_operation_command(self)

    def _get_shifts(self):
        """
        Get the mounting position from the Diffractometer DS.

        @return: 3-tuple
        """
        if self.shifts_channel is not None:
            shifts = self.shifts_channel.get_value()
        else:
            shifts = None
        return shifts


def test_hwo(hwo):
    print(hwo._get_shifts())



"""
CATS maintenance mockup.
"""

import logging
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task

__author__ = "Mikel Eguiraun"
__credits__ = ["The MxCuBE collaboration"]


TOOL_FLANGE, TOOL_UNIPUCK, TOOL_SPINE, TOOL_PLATE, TOOL_LASER, TOOL_DOUBLE_GRIPPER = (
    0,
    1,
    2,
    3,
    4,
    5,
)

TOOL_TO_STR = {
    "Flange": TOOL_FLANGE,
    "Unipuck": TOOL_UNIPUCK,
    "Rotat": TOOL_SPINE,
    "Plate": TOOL_PLATE,
    "Laser": TOOL_LASER,
    "Double": TOOL_DOUBLE_GRIPPER,
}


class CatsMaintMockup(HardwareObject):
    __TYPE__ = "CATS"
    NO_OF_LIDS = 3

    """
    Actual implementation of the CATS Sample Changer, MAINTENANCE COMMANDS ONLY
    BESSY BL14.1 installation with 3 lids
    """

    def __init__(self, name):
        super().__init__(name)

        self._state = "READY"
        self._running = 0
        self._powered = 0
        self._toolopen = 0
        self._message = "Nothing to report"
        self._regulating = 0
        self._lid1state = 0
        self._lid2state = 0
        self._lid3state = 0
        self._charging = 0
        self._currenttool = 1

    def init(self):
        try:
            self.cats_model = self.cats_device.read_attribute("CatsModel").value
        except Exception:
            self.cats_model = "CATS"

    def get_current_tool(self):
        return self._currenttool

    ################################################################################
    def _do_abort(self):
        """
        Launch the "abort" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        pass

    def _do_reset(self):
        """
        Launch the "reset" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        pass

    def _do_dry_gripper(self):
        """
        Launch the "dry" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        pass

    def _do_set_on_diff(self, sample):
        """
        Launch the "setondiff" command on the CATS Tango DS, an example of sample value is 2:05

        :returns: None
        :rtype: None
        """

        if sample is None:
            raise Exception("No sample selected")
        else:
            str_tmp = str(sample)
            sample_tmp = str_tmp.split(":")
            # calculate CATS specific lid/sample number
            lid = (int(sample_tmp[0]) - 1) / 3 + 1
            puc_pos = ((int(sample_tmp[0]) - 1) % 3) * 10 + int(sample_tmp[1])
            argin = [str(lid), str(puc_pos), "0"]
            logging.getLogger().info("to SetOnDiff %s", argin)
            # self._execute_server_task(self._cmdSetOnDiff,argin)

    def _do_power_state(self, state=False):
        """
        Switch on CATS power if >state< == True, power off otherwise

        :returns: None
        :rtype: None
        """
        self._powered = state
        self._update_powered_state(state)

    def _do_enable_regulation(self):
        """
        Switch on CATS regulation

        :returns: None
        :rtype: None
        """
        self._regulating = True
        self._update_regulation_state(True)

    def _do_disable_regulation(self):
        """
        Switch off CATS regulation

        :returns: None
        :rtype: None
        """
        self._regulating = False
        self._update_regulation_state(False)

    def _do_lid1_state(self, state=True):
        """
        Opens lid 1 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        self._lid1state = state
        self._update_lid1_state(state)

    def _do_lid2_state(self, state=True):
        """
        Opens lid 2 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        self._lid2state = state
        self._update_lid2_state(state)

    def _do_lid3_state(self, state=True):
        """
        Opens lid 3 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        self._lid3state = state
        self._update_lid3_state(state)

    #########################          PROTECTED          #########################

    def _execute_task(self, wait, method, *args):
        ret = self._run(method, *args)
        if wait:
            return ret.get()
        else:
            return ret

    @task
    def _run(self, method, *args):
        exception = None
        ret = None
        try:
            ret = method(*args)
        except Exception as ex:
            exception = ex
        if exception is not None:
            raise exception
        return ret

    #########################           PRIVATE           #########################

    def _update_running_state(self, value):
        self._running = value
        self.emit("runningStateChanged", (value,))
        self._update_global_state()

    def _update_powered_state(self, value):
        self._powered = value
        self.emit("powerStateChanged", (value,))
        self._update_global_state()

    def _update_tool_state(self, value):
        self._toolopen = value
        self.emit("toolStateChanged", (value,))
        self._update_global_state()

    def _update_message(self, value):
        self._message = value
        self.emit("messageChanged", (value,))
        self._update_global_state()

    def _update_regulation_state(self, value):
        self._regulating = value
        self.emit("regulationStateChanged", (value,))
        self._update_global_state()

    def _update_state(self, value):
        self._state = value
        self._update_global_state()

    def _update_lid1_state(self, value):
        self._lid1state = value
        self.emit("lid1StateChanged", (value,))
        self._update_global_state()

    def _update_lid2_state(self, value):
        self._lid2state = value
        self.emit("lid2StateChanged", (value,))
        self._update_global_state()

    def _update_lid3_state(self, value):
        self._lid3state = value
        self.emit("lid3StateChanged", (value,))
        self._update_global_state()

    def _update_operation_mode(self, value):
        self._charging = not value

    def _update_global_state(self):
        state_dict, cmd_state, message = self.get_global_state()
        self.emit("globalStateChanged", (state_dict, cmd_state, message))

    def get_global_state(self):
        """
        Update clients with a global state that
        contains different:

        - first param (state_dict):
            collection of state bits

        - second param (cmd_state):
            list of command identifiers and the
            status of each of them True/False
            representing whether the command is
            currently available or not

        - message
            a message describing current state information
            as a string
        """
        _ready = str(self._state) in ("READY", "ON")

        if self._running:
            state_str = "MOVING"
        elif not (self._powered) and _ready:
            state_str = "DISABLED"
        elif _ready:
            state_str = "READY"
        else:
            state_str = str(self._state)

        state_dict = {
            "toolopen": self._toolopen,
            "powered": self._powered,
            "running": self._running,
            "regulating": self._regulating,
            "lid1": self._lid1state,
            "lid2": self._lid2state,
            "lid3": self._lid3state,
            "state": state_str,
        }

        cmd_state = {
            "PowerOn": (not self._powered) and _ready,
            "PowerOff": (self._powered) and _ready,
            "regulon": (not self._regulating) and _ready,
            "openlid1": (not self._lid1state) and self._powered and _ready,
            "closelid1": self._lid1state and self._powered and _ready,
            "dry": (not self._running) and self._powered and _ready,
            "soak": (not self._running) and self._powered and _ready,
            "home": (not self._running) and self._powered and _ready,
            "back": (not self._running) and self._powered and _ready,
            "safe": (not self._running) and self._powered and _ready,
            "clear_memory": True,
            "reset": True,
            "abort": self._running,
        }

        message = self._message

        return state_dict, cmd_state, message

    def get_cmd_info(self):
        """return information about existing commands for this object
        the information is organized as a list
        with each element contains
        [ cmd_name,  display_name, category ]
        """
        """ [cmd_id, cmd_display_name, nb_args, cmd_category, description ] """
        cmd_list = [
            [
                "Power",
                [
                    ["PowerOn", "PowerOn", "Switch Power On"],
                    ["PowerOff", "PowerOff", "Switch Power Off"],
                    ["regulon", "Regulation On", "Swich LN2 Regulation On"],
                ],
            ],
            [
                "Lid",
                [
                    ["openlid1", "Open Lid", "Open Lid"],
                    ["closelid1", "Close Lid", "Close Lid"],
                ],
            ],
            [
                "Actions",
                [
                    ["home", "Home", "Actions", "Home (trajectory)"],
                    ["dry", "Dry", "Actions", "Dry (trajectory)"],
                    ["soak", "Soak", "Actions", "Soak (trajectory)"],
                ],
            ],
            [
                "Recovery",
                [
                    [
                        "clear_memory",
                        "Clear Memory",
                        "Clear Info in Robot Memory "
                        " (includes info about sample on Diffr)",
                    ],
                    ["reset", "Reset Message", "Reset Cats State"],
                    ["back", "Back", "Reset Cats State"],
                    ["safe", "Safe", "Reset Cats State"],
                ],
            ],
            ["Abort", [["abort", "Abort", "Abort Execution of Command"]]],
        ]
        return cmd_list

    def _execute_server_task(self, method, *args):
        task_id = method(*args)
        ret = None
        # introduced wait because it takes some time before the attribute PathRunning is set
        # after launching a transfer
        # after setting refresh in the Tango DS to 0.1 s a wait of 1s is enough
        time.sleep(1.0)
        while str(self._chnPathRunning.get_value()).lower() == "true":
            gevent.sleep(0.1)
        ret = True
        return ret

    def send_command(self, cmd_name, args=None):
        lid = 1
        toolcal = 0
        tool = self.get_current_tool()

        if cmd_name in ["dry", "safe", "home"]:
            if tool is not None:
                args = [tool]
            else:
                raise Exception("Cannot detect type of TOOL in Cats. Command ignored")

        if cmd_name == "soak":
            if tool in [TOOL_DOUBLE_GRIPPER, TOOL_UNIPUCK]:
                args = [str(tool), str(lid)]
            else:
                raise Exception("Can SOAK only when UNIPUCK tool is mounted")

        if cmd_name == "back":
            if tool is not None:
                args = [tool, toolcal]
            else:
                raise Exception("Cannot detect type of TOOL in Cats. Command ignored")

        if cmd_name == "PowerOn":
            self._do_power_state(True)
        if cmd_name == "PowerOff":
            self._do_power_state(False)

        if cmd_name == "regulon":
            self._do_enable_regulation()
        if cmd_name == "reguloff":
            self._do_disable_regulation()
        if cmd_name == "openlid1":
            self._do_lid1_state(True)
        if cmd_name == "closelid1":
            self._do_lid1_state(False)
        return True


def test_hwo(hwo):
    print((hwo.get_current_tool()))



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import time

from PyTango.gevent import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector

__author__ = "Vicente Rey"
__credits__ = ["ALBA"]
__version__ = "2.3."
__category__ = "General"


class ALBAPilatus(AbstractDetector, HardwareObject):
    """Detector class. Contains all information about detector
    - states are 'OK', and 'BAD'
    - status is busy, exposing, ready, etc.
    - physical property is RH for pilatus, P for rayonix
    """

    def __init__(self, name):
        AbstractDetector.__init__(self)
        HardwareObject.__init__(self, name)

        self.distance_motor_hwobj = None
        self.default_distance = None
        self.default_distance_limits = None

        self.default_latency_time = 0.003

        self.exp_time_limits = None

        self.headers = {}

    def init(self):
        self.distance_motor_hwobj = self.get_object_by_role("distance_motor")
        self.devname = self.get_property("tangoname")

        try:
            self.latency_time = float(self.get_property("latency_time"))
        except Exception:
            self.latency_time = None

        if self.latency_time is None:
            self.log.debug(
                "Cannot obtain latency time from Pilatus XML. Using %s"
                % self.default_latency_time
            )
            self.latency_time = self.default_latency_time

        self.devspecific = self.get_property("device_specific")

        exp_time_limits = self.get_property("exposure_limits")
        self.exp_time_limits = map(float, exp_time_limits.strip().split(","))

        self.device = DeviceProxy(self.devname)
        self.device_specific = DeviceProxy(self.devspecific)
        self.device.set_timeout_millis(30000)

        self.beamx_chan = self.get_channel_object("beamx")
        self.beamy_chan = self.get_channel_object("beamy")

    def start_acquisition(self):
        self.device.startAcq()

    def stop_acquisition(self):
        self.device.abortAcq()

    def wait_move_distance_done(self):
        self.distance_motor_hwobj.wait_end_of_move()

    def get_threshold(self):
        return self.device_specific.threshold

    def get_threshold_gain(self):
        return self.device_specific.threshold_gain

    def has_shutterless(self):
        """Return True if has shutterless mode"""
        return True

    def get_beam_position(self, distance=None, wavelength=None):
        """Returns beam center coordinates"""

        # NBNB TODO check if pixels or mm, and adjust code
        # Should be pixels
        beam_x = 0
        beam_y = 0
        try:
            beam_x = self.beamx_chan.get_value()
            beam_y = self.beamy_chan.get_value()
        except Exception:
            self.log.exception("")
        return beam_x, beam_y

    def get_manufacturer(self):
        return self.get_property("manufacturer")
        return "Dectris"

    def get_model(self):
        return self.get_property("model")

    def get_detector_type(self):
        return self.get_property("type")

    def get_default_exposure_time(self):
        return self.get_property("default_exposure_time")

    def get_minimum_exposure_time(self):
        return self.get_property("minimum_exposure_time")

    def get_exposure_time_limits(self):
        """Returns exposure time limits as list with two floats"""
        return self.exp_time_limits

    def get_file_suffix(self):
        return self.get_property("file_suffix")

    def get_pixel_size(self):
        return self.get_property("px"), self.get_property("py")

    # methods for data collection
    def set_energy_threshold(self):
        eugap_ch = self.get_channel_object("eugap")

        try:
            currentenergy = eugap_ch.get_value()
        except Exception:
            currentenergy = 12.6

        det_energy = self.get_threshold()

        # threshold = det_energy  / 2.
        # limitenergy = threshold / 0.8

        if round(currentenergy, 6) < 7.538:
            currentenergy = 7.538

        kev_diff = abs(det_energy - currentenergy)

        if kev_diff > 1.2:
            self.log.debug(
                "programming energy_threshold on pilatus to: %s" % currentenergy
            )
            # if self.wait_standby():
            # self.device_specific.energy_threshold = currentenergy

    def get_latency_time(self):
        return self.latency_time

    def wait_standby(self, timeout=300):
        t0 = time.time()
        while self.device_specific.cam_state == "STANDBY":
            if time.time() - t0 > timeout:
                print("timeout waiting for Pilatus to be on STANDBY")
                return False
            time.sleep(0.1)
        return True

    def prepare_acquisition(self, dcpars):
        self.set_energy_threshold()
        # self.wait_standby()

        osc_seq = dcpars["oscillation_sequence"][0]
        file_pars = dcpars["fileinfo"]

        basedir = file_pars["directory"]
        prefix = "%s_%s_" % (file_pars["prefix"], file_pars["run_number"])

        first_img_no = osc_seq["start_image_number"]
        nb_frames = osc_seq["number_of_images"]
        exp_time = osc_seq["exposure_time"]

        fileformat = "CBF"
        trig_mode = "EXTERNAL_TRIGGER"
        # latency_time = 0.003

        self.log.debug(
            " Preparing detector (dev=%s) for data collection" % self.devname
        )

        self.log.debug("    /saving directory: %s" % basedir)
        self.log.debug("    /prefix          : %s" % prefix)
        self.log.debug("    /saving_format   : %s" % fileformat)
        self.log.debug("    /trigger_mode    : %s" % trig_mode)
        self.log.debug("    /acq_nb_frames   : %s" % nb_frames)
        self.log.debug("    /acq_expo_time   : %s" % str(exp_time - self.latency_time))
        self.log.debug("    /latency_time    : %s" % self.latency_time)

        self.device.write_attribute("saving_mode", "AUTO_FRAME")
        self.device.write_attribute("saving_directory", basedir)
        self.device.write_attribute("saving_prefix", prefix)
        self.device.write_attribute("saving_format", fileformat)

        # set ROI and header in limaserver
        #  TODO

        TrigList = [
            "INTERNAL_TRIGGER",
            "EXTERNAL_TRIGGER",
            "EXTERNAL_TRIGGER_MULTI",
            "EXTERNAL_GATE",
            "EXTERNAL_START_STOP",
        ]

        self.device.write_attribute("acq_trigger_mode", trig_mode)
        self.device.write_attribute("acq_expo_time", exp_time - self.latency_time)
        self.device.write_attribute("latency_time", self.latency_time)

        return True

    def prepare_collection(self, nb_frames, first_img_no):
        self.log.debug(
            "ALBAPilatus. preparing collection. nb_images: %s, first_no: %s"
            % (nb_frames, first_img_no)
        )
        self.device.write_attribute("acq_nb_frames", nb_frames)
        self.device.write_attribute("saving_next_number", first_img_no)
        self.device.prepareAcq()
        return True

    def start_collection(self):
        self.start_acquisition()

    def stop_collection(self):
        self.stop_acquisition()

    def set_image_headers(self, image_headers, angle_info):
        nb_images = image_headers["nb_images"]
        angle_inc = image_headers["Angle_increment"]
        start_angle = image_headers["Start_angle"]

        startangles_list = []
        ang_start, ang_inc, spacing = angle_info
        for i in range(nb_images):
            startangles_list.append("%0.4f deg." % (ang_start + spacing * i))

        headers = []
        for i, sa in enumerate(startangles_list):
            header = (
                "_array_data.header_convention PILATUS_1.2\n"
                "# Detector: PILATUS 6M, S/N 60-0108, Alba\n"
                "# %s\n"
                "# Pixel_size 172e-6 m x 172e-6 m\n"
                "# Silicon sensor, thickness 0.000320 m\n"
                % time.strftime("%Y/%b/%d %T")
            )

            # Acquisition values (headers dictionary) but overwrites start angle
            image_headers["Start_angle"] = sa
            for key, value in image_headers.items():
                if key == "nb_images":
                    continue
                header += "# %s %s\n" % (key, value)
            headers.append("%d : array_data/header_contents|%s;" % (i, header))

        self.device.write_attribute("saving_header_delimiter", ["|", ";", ":"])
        self.device.resetCommonHeader()
        self.device.resetFrameHeaders()
        self.device.setImageHeader(headers)


def test_hwo(hwo):
    print("Detector Distance is: ", hwo.distance.get_value())
    print("   Beam X: %s / Beam Y: %s" % hwo.get_beam_position())
    # print("going to 490 : ", hwo.distance.set_value(490))



"""
The ALBAFastShutter hardware object is a variation of the ALBAEpsActuator
where the command to open/close is done on a different channel than the
reading of the shutter state.

The interface is otherwise exactly the same as the ALBAEpsActuator

Example XML::

  <object class="ALBAFrontEnd">
    <username>Front Shutter</username>
    <taurusname>bl13/ct/eps-plc-01</taurusname>
    <channel type="sardana" polling="events" name="actuator">fe_open</channel>
    <channel type="sardana" name="open_command">OPEN_FE</channel>
    <channel type="sardana" name="close_command">CLOSE_FE</channel>
    <states>Open,Closed</states>
  </object>

"""

from ALBAEpsActuator import ALBAEpsActuator


class ALBAFrontEnd(ALBAEpsActuator):
    def init(self):
        ALBAEpsActuator.init(self)

        self.open_channel = self.get_channel_object("open_command")
        self.close_channel = self.get_channel_object("close_command")

    def cmdIn(self):
        self.open_channel.set_value(True)
        # self.actuator_channel.set_value(1)

    def cmdOut(self):
        self.close_channel.set_value(True)
        # self.actuator_channel.set_value(0)


def test_hwo(hwo):
    print("Name is: ", hwo.getUserName())
    print("Shutter state is: ", hwo.get_state())
    print("Shutter status is: ", hwo.getStatus())

    # print "Opening it"
    # print hwo.open()
    # print "Closing it"
    # print hwo.close()



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Energy and Wavelength with bliss.
Example yaml file:
- for tunable wavelength beamline:
.. code-block:: yaml

 class: BlissEnergy.BlissEnergy
 configuration:
   username: Energy
 objects:
   controller: bliss.yml
   energy_motor: energy_motor.yml

- for fixed wavelength beamline:
.. code-block:: yaml

 class: BlissEnergy.BlissEnergy
 configuration:
   username: Energy
   read_only: True
   default_value: 12.8123

"""

import logging
import math

from gevent import spawn

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class BlissEnergy(AbstractEnergy):
    """Energy and Wavelength with bliss."""

    def __init__(self, name):
        super().__init__(name)
        self.energy_motor = None
        self.controller = None
        self._cmd_execution = None

    def init(self):
        """Initialisation"""
        super().init()

        self.controller = self.get_object_by_role("controller")
        self.energy_motor = self.get_object_by_role("energy_motor")
        self.update_state(HardwareObjectState.READY)

        if self.energy_motor:
            self.update_state(self.energy_motor.get_state())
            self.energy_motor.connect("valueChanged", self.update_value)
            self.energy_motor.connect("stateChanged", self.update_state)

        if self.read_only and not self.energy_motor:
            # self._nominal_value = float(self.get_property("energy", 0))
            try:
                self._nominal_value = float(self.default_value)
            except TypeError as err:
                msg = "Energy not defined"
                raise RuntimeError(msg) from err

    def get_value(self):
        """Read the energy.
        Returns:
            (float): Energy [keV]
        """
        if self.energy_motor:
            self._nominal_value = self.energy_motor.get_value()
        return self._nominal_value

    def get_limits(self):
        """Return energy low and high limits.
        Returns:
            (tuple): two floats tuple (low limit, high limit) [keV].
        """
        if not self.read_only:
            self._nominal_limits = self.energy_motor.get_limits()
        return self._nominal_limits

    def stop(self):
        """Stop the energy motor movement"""
        self.energy_motor.stop()

    def _set_value(self, value):
        """Execute the sequence to move to an energy
        Args:
            value (float): target energy
        """
        try:
            defocus = HWR.beamline.beam.definer.defocused_beam
        except AttributeError:
            defocus = False

        self.update_state(HardwareObjectState.BUSY)
        try:
            self.controller.change_energy(value, defocus=defocus)
        except (AttributeError, RuntimeError):
            self.energy_motor.set_value(value)
        self.update_state(HardwareObjectState.READY)

    def set_value(self, value, timeout=0):
        """Move energy to absolute position. Wait the move to finish.
        Args:
            value (float): target value.
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait
                             if timeout is None: wait forever.
        Raises:
            ValueError: Value not valid or attempt to set write only actuator.
        """
        if self.read_only:
            return

        if self.validate_value(value):
            current_value = self.get_value()

            _delta = math.fabs(current_value - value)
            if _delta < 0.001:
                logging.getLogger("user_level_log").info(
                    "Energy: already at %g, not moving", value
                )
                return

            logging.getLogger("user_level_log").info(
                "Energy: moving energy to %g", value
            )

            if _delta > 0.02:
                if timeout:
                    self._set_value(value)
                else:
                    self._cmd_execution = spawn(self._set_value, value)
            else:
                self.energy_motor.set_value(value, timeout=timeout)
        else:
            msg = f"Invalid value {value}"
            raise ValueError(msg)

    def abort(self):
        """Abort the procedure"""
        if self._cmd_execution and not self._cmd_execution.ready():
            self._cmd_execution.kill()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


import logging
import xml.sax
from xml.sax.handler import ContentHandler

from mxcubecore import BaseHardwareObjects

CURRENT_XML = None

new_objects_classes = {
    "equipment": BaseHardwareObjects.Equipment,
    "device": BaseHardwareObjects.Device,
    "procedure": BaseHardwareObjects.Procedure,
}


def parse(filename, name):
    """[summary]

    Args:
        filename ([type]): [description]
        name ([type]): [description]

    Returns:
        [type]: [description]
    """
    cur_handler = HardwareObjectHandler(name)

    global CURRENT_XML
    try:
        xml_file = open(filename)
        CURRENT_XML = xml_file.read()
    except Exception:
        CURRENT_XML = None

    xml.sax.parse(filename, cur_handler)

    return cur_handler.get_hardware_object()


def parse_string(xml_hardware_object, name):
    """[summary]

    Args:
        xml_hardware_object ([type]): [description]
        name ([type]): [description]

    Returns:
        [type]: [description]
    """
    global CURRENT_XML
    CURRENT_XML = xml_hardware_object
    cur_handler = HardwareObjectHandler(name)
    xml.sax.parseString(str.encode(xml_hardware_object), cur_handler)
    return cur_handler.get_hardware_object()


def load_module(hardware_object_name):
    """[summary]

    Args:
        hardware_object_name ([type]): [description]

    Returns:
        [type]: [description]
    """
    return __import__(hardware_object_name, globals(), locals(), [""])


def instanciate_class(module_name, class_name, object_name):
    """[summary]

    Args:
        module_name ([type]): [description]
        class_name ([type]): [description]
        object_name ([type]): [description]

    Returns:
        [type]: [description]
    """
    module = load_module(module_name)
    if module is None:
        return
    else:
        try:
            class_obj = getattr(module, class_name)
        except AttributeError:
            logging.getLogger("HWR").error(
                "No class %s in module %s", class_name, module_name
            )
        else:
            # check the XML
            if module.__doc__ is not None and CURRENT_XML is not None:
                i = module.__doc__.find("template:")

                if i >= 0:
                    xml_template = module.__doc__[i + 10 :]

                    xml_structure_retriever = XmlStructureRetriever()
                    xml.sax.parseString(CURRENT_XML, xml_structure_retriever)
                    current_structure = xml_structure_retriever.get_structure()
                    xml_structure_retriever = XmlStructureRetriever()
                    xml.sax.parseString(xml_template, xml_structure_retriever)
                    template_structure = xml_structure_retriever.get_structure()

                    if not template_structure == current_structure:
                        logging.getLogger("HWR").error(
                            "%s: XML file does not match the %s class template"
                            % (object_name, class_name)
                        )
                        return
            try:
                new_instance = class_obj(object_name)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Cannot instanciate class %s", class_name
                )
            else:
                return new_instance


class HardwareObjectHandler(ContentHandler):
    def __init__(self, name):
        """[summary]

        Args:
            name ([type]): [description]
        """
        ContentHandler.__init__(self)

        self.name = name
        self.class_error = False
        self.objects = []
        self.reference = ""
        self.property = ""
        self.element_is_a_reference = False
        self.element_role = None
        self.buffer = ""
        self.path = ""
        self.previous_path = ""
        self.hwr_import_reference = None

    def get_hardware_object(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        if self.hwr_import_reference is not None:
            return self.hwr_import_reference
        elif len(self.objects) == 1:
            return self.objects[0]

    def startElement(self, name, attrs):
        """[summary]

        Args:
            name ([type]): [description]
            attrs ([type]): [description]
        """
        if self.class_error:
            return

        self.buffer = ""

        if len(self.objects) == 0:
            object_name = self.name
        else:
            object_name = name

        assert not self.element_is_a_reference

        self.element_role = None
        self.property = ""
        self.command = {}
        self.channel = {}

        #
        # determine path to the new object
        #
        self.path += "/" + str(name) + "[%d]"
        i = self.previous_path.rfind("[")

        if i >= 0 and self.path[:-4] == self.previous_path[:i]:
            object_index = int(self.previous_path[i + 1 : -1]) + 1
        else:
            object_index = 1  # XPath indexes begin at 1

        self.path %= object_index

        _attrs = attrs
        attrs = {}

        for k in list(_attrs.keys()):
            v = str(_attrs[k])

            if v == "None":
                attrs[str(k)] = None
            else:
                try:
                    attrs[str(k)] = int(v)
                except Exception:
                    try:
                        attrs[str(k)] = float(v)
                    except Exception:
                        if v == "False":
                            attrs[str(k)] = False
                        elif v == "True":
                            attrs[str(k)] = True
                        else:
                            attrs[str(k)] = v
        if name == "hwr_import":
            self.hwr_import_reference = attrs["href"]

        if "role" in attrs:
            self.element_role = attrs["role"]
        if name == "device":
            # maybe we have to add the DeviceContainer mix-in class to each node of
            # the Hardware Object hierarchy
            i = len(self.objects) - 1
            while i >= 0 and not isinstance(
                self.objects[i], BaseHardwareObjects.DeviceContainer
            ):
                # newClass = new.classobj("toto", (self.objects[i].__class__,) + self.objects[i].__class__.__bases__ + (BaseHardwareObjects.DeviceContainer, ), {})
                # TODO replace deprecated DeviceContainerNode with a different class
                self.objects[i].__class__ = BaseHardwareObjects.DeviceContainerNode
                i -= 1

        #
        # is element a reference to another hardware object ?
        #
        ref = "hwrid" in attrs and attrs["hwrid"] or "href" in attrs and attrs["href"]
        if ref:
            self.element_is_a_reference = True
            self.reference = str(ref)

            if self.reference.startswith("../"):
                self.reference = "/".join(
                    self.name.split("/")[:-1] + [self.reference[3:]]
                )
            elif self.reference.startswith("./"):
                self.reference = "/".join(
                    self.name.split("/")[:-1] + [self.reference[2:]]
                )
            return

        if name in new_objects_classes:
            if "class" in attrs:
                module_name = str(attrs["class"])
                class_name = module_name.split(".")[-1]

                new_object = instanciate_class(module_name, class_name, object_name)

                if new_object is None:
                    self.class_error = True
                    return
                else:
                    new_object.set_path(self.path)
                    self.objects.append(new_object)
            else:
                new_object_class = new_objects_classes[name]
                new_object = new_object_class(object_name)
                new_object.set_path(self.path)

                self.objects.append(new_object)
        elif name == "command":
            if "name" in attrs and "type" in attrs:
                # short command notation
                self.command.update(attrs)
            else:
                # long command notation (allow arguments)
                self.objects.append(BaseHardwareObjects.HardwareObjectNode(object_name))
        elif name == "channel":
            if "name" in attrs and "type" in attrs:
                self.channel.update(attrs)
        else:
            if len(self.objects) == 0:
                if "class" in attrs:
                    module_name = str(attrs["class"])
                    class_name = module_name.split(".")[-1]

                    new_object = instanciate_class(module_name, class_name, object_name)

                    if new_object is None:
                        self.class_error = True
                        return
                else:
                    new_object = BaseHardwareObjects.HardwareObject(object_name)

                new_object.set_path(self.path)
                self.objects.append(new_object)
                """
                # maybe we can create a HardwareObject ? be strict for the moment...
                logging.getLogger("HWR").error("%s: unknown Hardware Object type (should be one of %s)", object_name, str(new_objects_classes.keys()))
                self.class_error = True
                return
                """
            else:
                new_object = BaseHardwareObjects.HardwareObjectNode(object_name)
                new_object.set_path(self.path)
                self.objects.append(new_object)

                self.property = name  # element is supposed to be a Property

    def characters(self, content):
        """[summary]

        Args:
            content ([type]): [description]
        """
        if self.class_error:
            return

        self.buffer += str(content)

    def endElement(self, name):
        """[summary]

        Args:
            name ([type]): [description]
        """
        if self.class_error:
            return

        name = str(name)

        if self.element_is_a_reference:
            if len(self.objects) > 0:
                self.objects[-1].add_reference(
                    name, self.reference, role=self.element_role
                )
                self.objects[0].add_reference(
                    name, self.reference, role=self.element_role
                )
        else:
            try:
                if name == "command":
                    if len(self.command) > 0:
                        if len(self.objects) > 0:
                            self.objects[-1].add_command(
                                self.command, self.buffer, add_now=False
                            )
                    else:
                        if len(self.objects) > 1:
                            self.objects[-2].add_command(
                                self.objects.pop(), add_now=False
                            )
                elif name == "channel":
                    if len(self.channel) > 0:
                        if len(self.objects) > 0:
                            self.objects[-1].add_channel(
                                self.channel, self.buffer, add_now=False
                            )
                elif name == self.property:
                    del self.objects[-1]  # remove empty object
                    self.objects[-1]._set_property(name, self.buffer)  # noqa: SLF001
                else:
                    if len(self.objects) == 1:
                        return

                    if len(self.objects) > 1:
                        self.objects[-2]._add_object(  # noqa: SLF001
                            name, self.objects[-1], role=self.element_role
                        )
                    if len(self.objects) > 0:
                        del self.objects[-1]
            except Exception:
                logging.getLogger("HWR").exception(
                    "%s: error while creating Hardware Object from XML file", self.name
                )

        self.element_is_a_reference = False
        self.element_role = None
        self.buffer = ""
        self.previous_path = self.path
        self.path = self.path[
            : self.path.rfind("/")
        ]  # remove last added name and suffix


class XMLStructure:
    def __init__(self):
        self.xmlpaths = set()
        self.attributes = {}

    def add(self, xml_path, attributes_set):
        """[summary]

        Args:
            xml_path ([type]): [description]
            attributes_set ([type]): [description]
        """
        self.xmlpaths.add(xml_path)

        if len(attributes_set) > 0:
            self.attributes[xml_path] = attributes_set

    def __eq__(self, s):
        """[summary]

        Args:
            s ([type]): [description]

        Returns:
            [type]: [description]
        """
        if self.xmlpaths.issubset(s.xmlpaths):
            for xml_path, attribute_set in self.attributes.items():
                try:
                    attribute_set_2 = s.attributes[xml_path]
                except KeyError:
                    return False
                else:
                    if not attribute_set.issubset(attribute_set_2):
                        return False

            return True
        else:
            return False


class XmlStructureRetriever(ContentHandler):
    def __init__(self):
        ContentHandler.__init__(self)

        self.path = ""
        self.previous_path = ""
        self.current_attributes = set()
        self.structure = XMLStructure()

    def get_structure(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        return self.structure

    def startElement(self, name, attrs):
        """[summary]

        Args:
            name ([type]): [description]
            attrs ([type]): [description]
        """
        self.path += "/" + str(name) + "[%d]"
        i = self.previous_path.rfind("[")

        if i >= 0 and self.path[:-4] == self.previous_path[:i]:
            index = int(self.previous_path[i + 1 : -1]) + 1
        else:
            index = 1  # XPath indexes begin at 1

        self.path %= index

        for attr, value in list(attrs.items()):
            if str(attr) == "hwrid":
                attr = "href"

            self.current_attributes.add("%s=%s" % (str(attr), str(value)))

    def endElement(self, name):
        """[summary]

        Args:
            name ([type]): [description]
        """
        self.structure.add(self.path, self.current_attributes)

        self.previous_path = self.path
        self.path = self.path[: self.path.rfind("/")]



from mxcubecore.BaseHardwareObjects import HardwareObject


class ALBASupervisor(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)

    def init(self):
        self.state_chan = self.get_channel_object("state")
        self.go_collect_cmd = self.get_command_object("go_collect")
        self.go_sample_view_cmd = self.get_command_object("go_sample_view")
        self.go_transfer_cmd = self.get_command_object("go_transfer")
        self.go_beam_view_cmd = self.get_command_object("go_beam_view")

        self.phase_chan = self.get_channel_object("phase")
        self.detector_cover_chan = self.get_channel_object("detector_cover_open")

        self.state_chan.connect_signal("update", self.state_changed)
        self.phase_chan.connect_signal("update", self.phase_changed)
        self.detector_cover_chan.connect_signal("update", self.detector_cover_changed)

    def is_ready(self):
        return True

    def getUserName(self):
        return self.username

    def state_changed(self, value):
        self.current_state = value
        self.emit("stateChanged", self.current_state)

    def phase_changed(self, value):
        if value == "Sample":
            value = "Centring"
        self.current_phase = value
        self.emit("phaseChanged", self.current_phase)

    def get_current_phase(self):
        return self.phase_chan.get_value()

    def go_collect(self):
        return self.go_collect_cmd()

    def go_transfer(self):
        return self.go_transfer_cmd()

    def go_sample_view(self):
        return self.go_sample_view_cmd()

    def go_beam_view(self):
        return self.go_beam_view_cmd()

    def get_state(self):
        return self.state_chan.get_value()

    def detector_cover_changed(self, value):
        self.detector_cover_opened = value
        # self.emit('levelChanged', self.current_level)

    def open_detector_cover(self):
        self.detector_cover_chan.set_value(True)

    def close_detector_cover(self, value):
        self.detector_cover_chan.set_value(False)

    def is_detector_cover_opened(self):
        return self.detector_cover_chan.get_value()


def test_hwo(hwo):
    print('\nSupervisor control "%s"\n' % hwo.getUserName())
    print("   Detector Cover  opened:", hwo.is_detector_cover_opened())
    print("   Current Phase is:", hwo.get_current_phase())
    print("   Current State is:", str(hwo.get_state()))



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""FlexHarvester Linux Java implementation of the Flex sample changer for Harvester Use.
Example xml file:
<object class = "EMBLFlexHarvester">
  <username>Sample Changer</username>
  <exporter_address>lid231flex1:9001</exporter_address>
</object>
"""

from __future__ import annotations

import logging
import time
from typing import (
    Any,
    List,
)

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.EMBLFlexHCD import EMBLFlexHCD
from mxcubecore.model import queue_model_objects as qmo
from mxcubecore.queue_entry.base_queue_entry import (
    CENTRING_METHOD,
    QueueExecutionException,
)
from mxcubecore.TaskUtils import task


class EMBLFlexHarvester(EMBLFlexHCD):
    """EMBLFlexHarvester is the Hardware Object interface for the EMBL Flex Sample Changer
    It inherits from EMBLFlexHCD and implements the Harvester interface.
    """

    __TYPE__ = "Flex Sample Changer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._harvester_hwo = None
        self.pin_cleaning = None
        self._loaded_sample = None

    def init(self):
        self.pin_cleaning = self.get_property("pin_cleaning")

        self._loaded_sample = (-1, -1, -1)
        self._harvester_hwo = self.get_object_by_role("harvester")

        EMBLFlexHCD.init(self)

    def get_room_temperature_mode(self) -> bool:
        """Get the Harvester Room Temperature Mode"""
        return self._execute_cmd_exporter("getRoomTemperatureMode", attribute=True)

    def set_room_temperature_mode(self, value) -> bool:
        """Set the Harvester Room Temperature Mode"""
        self._execute_cmd_exporter("setRoomTemperatureMode", value, command=True)
        logging.getLogger("user_level_log").info(
            "setting Robot Room temperature to %s", value
        )
        return self.get_room_temperature_mode()

    def mount_from_harvester(self):
        """This enable Mounting a sample from Harvester"""
        return True

    def get_sample_list(self) -> List[Any]:
        """
        Get Sample List related to the Harvester content/processing Plan
        """
        sample_list = super().get_sample_list()
        present_sample_list = []

        # Retrieve metadata lists from harvester HWO
        try:
            ha_sample_lists = self._harvester_hwo.get_crystal_uuids()
            ha_sample_names = self._harvester_hwo.get_sample_names()
            ha_sample_states = self._harvester_hwo.get_samples_state()
        except Exception as e:
            self.user_log.error(
                "Failed retrieving sample metadata from Harvester: %s", e
            )
            return present_sample_list

        ha_sample_acronyms = self._harvester_hwo.get_sample_acronyms()

        # If no samples reported by Harvester
        if not ha_sample_lists:
            self.user_log.warning("No samples reported by Harvester")
            return present_sample_list

        # Process each sample independently to avoid global failure
        for index, x_tal_uuid in enumerate(ha_sample_lists):
            sample = sample_list[index]
            img_url = self._harvester_hwo.get_crystal_images_urls(x_tal_uuid)
            img_target_x = self._harvester_hwo.get_image_target_x(x_tal_uuid)
            img_target_y = self._harvester_hwo.get_image_target_y(x_tal_uuid)

            # Enrich sample object in MXCuBE
            sample.id = x_tal_uuid
            sample.container_info = {"state": ha_sample_states[index]}

            sample._set_image_url(img_url)
            sample._set_image_x(img_target_x)
            sample._set_image_y(img_target_y)
            sample._name = ha_sample_names[index]

            # Protein acronym logic (fallback-safe)
            if ha_sample_acronyms and len(ha_sample_acronyms) == len(ha_sample_lists):
                sample.proteinAcronym = ha_sample_acronyms[index]
            else:
                sample.proteinAcronym = (
                    ha_sample_acronyms[0] if ha_sample_acronyms else ""
                )

            present_sample_list.append(sample)

        self.user_log.info("Loaded %d samples from Harvester", len(present_sample_list))
        return present_sample_list

    def _hw_get_mounted_sample(self) -> str:
        """Get the currently mounted sample location string
        Exporter Does not know the loaded sample location
        we get it from our own tracking
        """
        loaded_sample = self._loaded_sample
        return (
            str(loaded_sample[0])
            + ":"
            + str(loaded_sample[1])
            + ":"
            + "%02d" % loaded_sample[2]
        )

    def _hw_get_mounted_crystal_id(self) -> str:
        """Get the currently mounted crystal UUID"""
        return self._execute_cmd_exporter("getMountedCrystalId", attribute=True)

    @task
    def load_a_pin_for_calibration(self):
        """
        Load a Pin from Harvester
        """
        try:
            self.prepare_load()
            self.enable_power()

            load_task = gevent.spawn(
                self._execute_cmd_exporter,
                "loadSampleFromHarvester",
                self.pin_cleaning,
                command=True,
            )

            self._wait_busy(30)
            err_msg = "Timeout while waiting to sample to be loaded"
            with gevent.Timeout(600, RuntimeError(err_msg)):
                while not load_task.ready():
                    logging.getLogger("user_level_log").info("wait loading task")
                    gevent.sleep(2)

            with gevent.Timeout(600, RuntimeError(err_msg)):
                while True:
                    logging.getLogger("user_level_log").info("Wait Robot Safe position")
                    is_safe = self._execute_cmd_exporter(
                        "getRobotIsSafe", attribute=True
                    )
                    if is_safe:
                        break
                    gevent.sleep(2)
            return True
        except RuntimeError:
            return False

    def start_harvester_centring(self):
        try:
            dm = HWR.beamline.diffractometer

            logging.getLogger("user_level_log").info("Start Auto Harvesting Centring")

            computed_offset = HWR.beamline.harvester.get_offsets_for_sample_centering()
            dm.start_harvester_centring(computed_offset)

        except Exception as exc:
            logging.getLogger("user_level_log").exception(
                "Could not center sample, skipping"
            )
            raise QueueExecutionException(
                "Could not center sample, skipping", self
            ) from exc

    def _set_loaded_sample_and_prepare(self, loaded_sample_tup, previous_sample_tup):
        res = False

        loaded_sample = self.get_sample_with_address(loaded_sample_tup)

        if -1 not in loaded_sample_tup and loaded_sample_tup != previous_sample_tup:
            self._set_loaded_sample(loaded_sample)
            self._prepare_centring_task()
            res = True

        if res:
            if not self._harvester_hwo.get_room_temperature_mode():
                self.queue_harvest_next_sample(loaded_sample.get_address())

            # we expect CENTRING_METHOD to be None
            # NB: move this call to base_queue_entry mount_sample and add Harvester Centering METHOD
            HWR.beamline.queue_manager.centring_method = CENTRING_METHOD.NONE
            self.start_harvester_centring()

        return res

    def _do_load(self, sample=None) -> bool:
        """
        Load a Sample from Harvester
        """
        # 1. Initial setup and harvesting
        harvesting_result = self.queue_harvest_sample(sample.get_address())
        if not harvesting_result:
            raise QueueExecutionException("Harvester could not Harvest sample", self)

        self._update_state()
        self._wait_ready(600)

        sample_uuid = self.get_sample_uuid(sample.get_address())
        previous_sample = self._loaded_sample

        # 2. Main loading sequence
        if not self._perform_loading_sequence(sample_uuid):
            return False

        # 3 Finalization
        loaded_sample = (
            sample.get_cell_no(),
            sample.get_basket_no(),
            sample.get_vial_no(),
        )
        # Flex/Exporter Does not know the loaded sample location we keep track of it ourselfs
        self._loaded_sample = loaded_sample
        return self._set_loaded_sample_and_prepare(loaded_sample, previous_sample)

    def _perform_loading_sequence(self, sample_uuid) -> bool:
        """Start loading from harvester"""
        logging.getLogger("user_level_log").info(
            "Start loading from harvester SAMPLE_UUID %s", sample_uuid
        )
        load_task = gevent.spawn(
            self._execute_cmd_exporter,
            sample_uuid,
            self.pin_cleaning,
            command=True,
        )

        # Wait for sample changer to start activity
        try:
            _tt = time.time()
            self._wait_busy(300)
            logging.getLogger("HWR").info("Waited SC activity %s", time.time() - _tt)
        except Exception:
            logging.getLogger("user_level_log").error(
                "ERROR While Waited SC activity to start"
            )
            for msg in self.get_robot_exceptions():
                logging.getLogger("user_level_log").error(msg)
            raise

        #  Wait for sample to be loaded
        err_msg = "Timeout while waiting to sample to be loaded"
        try:
            with gevent.Timeout(600, RuntimeError(err_msg)):
                while not load_task.ready():
                    logging.getLogger("user_level_log").info("Wait loading task")
                    loaded_sample = self._hw_get_mounted_crystal_id()
                    if loaded_sample == sample_uuid:
                        break
                    gevent.sleep(2)
        except RuntimeError:
            logging.getLogger("user_level_log").error(err_msg)
            return False

        return self._finalize_loading_sequence()

    def _finalize_loading_sequence(self) -> bool:
        """Finalize loading sequence"""
        #  Wait for robot to be in safe state
        err_msg = "Timeout while waiting for robot to be in safe state"
        try:
            with gevent.Timeout(600, RuntimeError(err_msg)):
                while True:
                    is_safe = self._execute_cmd_exporter(
                        "getRobotIsSafe", attribute=True
                    )
                    if is_safe:
                        break
                    gevent.sleep(2)
        except RuntimeError:
            logging.getLogger("user_level_log").error(err_msg)
            return False

        # Check for robot exceptions
        for msg in self.get_robot_exceptions():
            if msg is not None and "Pin Cleaning Station" not in msg:
                logging.getLogger("user_level_log").error(
                    "ERROR While SC activity After Loaded Sample"
                )
                logging.getLogger("HWR").error(msg)
                logging.getLogger("user_level_log").error(msg)
                return False

        return True

    def harvest_and_mount_sample(self, xtal_uuid: str, sample) -> bool:
        """
        return (Bool) : whether the crystal has been Harvest then mount

        """
        try:
            self._harvester_hwo.harvest_crystal(xtal_uuid)
            self._harvester_hwo._wait_sample_transfer_ready(None)

            self._do_load(sample)
        except Exception:
            logging.getLogger("user_level_log").exception("Could not Harvest Crystal")
            return "Could not Harvest Crystal"

    def queue_list(self) -> List[str]:
        """
        builds a List representation of the queue based.
        """

        node = HWR.beamline.queue_model.get_model_root()

        result = []

        if isinstance(node, List):
            node_list = node
        else:
            node_list = node.get_children()

        for node in node_list:
            if isinstance(node, qmo.Sample):
                result.append(node.loc_str)

        return result

    def get_sample_uuid(self, sample_loc_str: str) -> str | None:
        """
        Get Sample UUID of the sample_loc_str from Sample List
        """
        samples_list = self.get_sample_list()
        sample_uuid = None
        for s in samples_list:
            # it seems get_id() was broken
            if s.get_address() == sample_loc_str or s.id == sample_loc_str:
                sample_uuid = s.id

                return sample_uuid
        return sample_uuid

    def queue_harvest_sample(self, sample_loc_str: str) -> bool:
        """
        While queue execution send harvest request
        """
        current_queue = self.queue_list()

        sample_uuid = self.get_sample_uuid(sample_loc_str)

        return self._harvester_hwo.queue_harvest_sample(
            sample_loc_str, sample_uuid, current_queue
        )

    def queue_harvest_next_sample(self, sample_loc_str) -> None:
        """
        While queue execution send harvest request
        on next sample of the queue list
        """

        current_queue_list = self.queue_list()

        next_sample = None
        try:
            next_sample = current_queue_list[
                current_queue_list.index(sample_loc_str) + 1
            ]
        except (ValueError, IndexError):
            next_sample = None

        sample_uuid = self.get_sample_uuid(next_sample)

        self._harvester_hwo.queue_harvest_next_sample(next_sample, sample_uuid)



from PyTransmission import matt_control

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class Transmission(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.labels = []
        self.indexes = []
        self.attno = 0
        # TO DO: clean this!!!
        # self.get_value = self.get_value
        # self.getAttFactor = self.get_value
        # self.setTransmission = self.set_value

    def init(self):
        self.__matt = matt_control.MattControl(
            self.get_property("wago_ip"),
            len(self["filter"]),
            0,
            self.get_property("type"),
            self.get_property("alternate"),
            self.get_property("status_module"),
            self.get_property("control_module"),
            self.get_property("datafile"),
        )
        self.__matt.connect()

    def is_ready(self):
        return True

    def getAtteConfig(self):
        self.attno = len(self["filter"])

        for att_i in range(self.attno):
            obj = self["filter"][att_i]
            self.labels.append(obj.label)
            self.indexes.append(obj.index)

    def getAttState(self):
        return self.__matt.pos_read()

    def _set_value(self, value):
        self.__matt.set_energy(HWR.beamline.energy.get_value())
        self.__matt.transmission_set(value)
        self._update()

    def _update(self):
        self.emit("attStateChanged", self.getAttState())
        self.emit("attFactorChanged", self.get_value())

    def toggle(self, attenuator_index):
        idx = self.indexes[attenuator_index]
        if self.is_in(attenuator_index):
            self.__matt.mattout(idx)
        else:
            self.__matt.mattin(idx)
        self._update()

    def get_value(self):
        self.__matt.set_energy(HWR.beamline.energy.get_value())
        return self.__matt.transmission_get()

    def is_in(self, attenuator_index):
        curr_bits = self.getAttState()
        idx = self.indexes[attenuator_index]
        return bool((1 << idx) & curr_bits)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import errno
import logging
import math
import os
import socket
import subprocess
import sys
import time

import gevent
import h5py
import numpy as np
import psutil

from mxcubecore import HardwareRepository as HWR
from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

FILE_TIMEOUT = 5


class P11Collect(AbstractCollect):
    def __init__(self, *args):
        super().__init__(*args)

    def init(self):
        """Initializes beamline collection parameters like default speed and server names."""
        super().init()
        self.default_speed = 120  # self.get_property("omega_default_speed", 130)
        self.turnback_time = 0.5  # self.get_property("turnback_time", 0.3)
        self.filter_server_name = self.get_property("filterserver")
        self.mono_server_name = self.get_property("monoserver")
        self.filter_server = DeviceProxy(self.filter_server_name)
        self.mono_server = DeviceProxy(self.mono_server_name)

        self.latest_frames = None
        self.acq_speed = None
        self.total_angle_range = None

    @task
    def move_motors(self, motor_position_dict):
        """Moves motors to the specified positions.

        Args:
            motor_position_dict (dict): Dictionary containing motor positions.
        """
        HWR.beamline.diffractometer.move_motors(motor_position_dict)

    def _take_crystal_snapshot(self, filename):
        """Takes a snapshot of the crystal and saves it to the given filename.

        Args:
            filename (str): Path to save the crystal snapshot.

        Raises:
            RuntimeError: If unable to move to the centring phase.
        """
        self.log.debug("#COLLECT# taking crystal snapshot.")

        if not HWR.beamline.diffractometer.is_centring_phase():
            self.log.debug("#COLLECT# take_snapshot. moving to centring phase")
            HWR.beamline.diffractometer.goto_centring_phase(wait=True)

        time.sleep(0.3)
        if not HWR.beamline.diffractometer.is_centring_phase():
            raise RuntimeError(
                "P11Collect. cannot reach centring phase for acquiring snapshots"
            )

        self.log.debug("#COLLECT# saving snapshot to %s" % filename)
        HWR.beamline.sample_view.save_snapshot(filename)

    def set_transmission(self, value):
        """Sets the transmission value on the beamline.

        Args:
            value (float): Transmission value to set.
        """
        HWR.beamline.transmission.set_value(value)

    def set_energy(self, value):
        """Sets the energy value on the beamline.

        Args:
            value (float): Energy value to set.
        """
        HWR.beamline.energy.set_value(value)

    def set_resolution(self, value):
        """Sets the resolution of the beamline.

        Args:
            value (float): Resolution value to set.
        """
        if not math.isclose(HWR.beamline.resolution.get_value(), value, abs_tol=0.01):
            HWR.beamline.resolution.set_value(value)

    def data_collection_hook(self):
        """Handles site-specific data collection processes."""
        dc_pars = self.current_dc_parameters
        collection_type = dc_pars["experiment_type"]

        self.log.debug(
            "======================= P11Collect. DATA COLLECTION HOOK =========================================="
        )
        self.log.debug(str(collection_type))
        self.log.debug(str(self.current_dc_parameters))

        osc_pars = self.current_dc_parameters["oscillation_sequence"][0]
        file_info = self.current_dc_parameters["fileinfo"]
        osc_pars["kappa"] = 0
        osc_pars["kappa_phi"] = 0

        start_angle = osc_pars["start"]
        img_range = osc_pars["range"]
        nframes = osc_pars["number_of_images"]
        self.latest_frames = nframes
        stop_angle = start_angle + img_range * nframes

        img_range = osc_pars["range"]
        exp_time = osc_pars["exposure_time"]
        self.acq_speed = img_range / exp_time

        self.total_angle_range = abs(stop_angle - start_angle)

        if not self.diffractometer_prepare_collection():
            raise RuntimeError("Cannot prepare diffractometer for collection")

        self.emit("progressStep", 5)

        try:
            self.log.debug("############# #COLLECT# Opening detector cover")
            HWR.beamline.diffractometer.detector_cover_open(wait=True)
            self.log.debug(
                "############ #COLLECT# detector cover is now open. Wait 2 more seconds"
            )
            time.sleep(2.0)  # wait extra time to allow det cover to be opened.

            basepath = file_info["directory"]
            prefix = file_info["prefix"]
            runno = file_info["run_number"]

            self.log.debug("#COLLECT# Programming detector for data collection")
            if collection_type == "Characterization":
                # Prepares metadata taken from current osc parameters.
                self.prepare_characterization()
                self.log.info(
                    "Collection: Creating directories for raw images and processing files EDNA and MOSFLM"
                )
                self.create_characterisation_directories()
                # Filepath for the presented data to work
                filepath = os.path.join(
                    basepath,
                    prefix,
                    "screening_"
                    + str(runno).zfill(3)
                    + "/"
                    + "%s_%d" % (prefix, runno),
                )

                self.log.debug(
                    "======= CURRENT FILEPATH: "
                    + str(filepath)
                    + "======================================="
                )
                self.latest_h5_filename = "%s_master.h5" % filepath
                self.log.debug(
                    "======= LATEST H5 FILENAME FILEPATH: "
                    + str(self.latest_h5_filename)
                    + "======================================="
                )

                self.log.debug(
                    "======= CURRENT FILEPATH: "
                    + str(filepath)
                    + "======================================="
                )
                self.latest_h5_filename = "%s_master.h5" % filepath
                self.log.debug(
                    "======= LATEST H5 FILENAME FILEPATH: "
                    + str(self.latest_h5_filename)
                    + "======================================="
                )

                angle_inc = 90.0
                HWR.beamline.detector.prepare_characterisation(
                    exp_time, nframes, angle_inc, filepath
                )

            else:
                # Prepares metadata taken from current osc parameters.
                self.prepare_std_collection(start_angle, img_range)
                self.log.info(
                    "Collection: Creating directories for raw images and processing files"
                )
                self.create_file_directories()

                filepath = os.path.join(
                    basepath,
                    prefix,
                    "rotational_"
                    + str(runno).zfill(3)
                    + "/"
                    + "%s_%d" % (prefix, runno),
                )

                self.log.debug(
                    "======= CURRENT FILEPATH: "
                    + str(filepath)
                    + "======================================="
                )
                self.latest_h5_filename = "%s_master.h5" % filepath
                self.log.debug(
                    "======= LATEST H5 FILENAME FILEPATH: "
                    + str(self.latest_h5_filename)
                    + "======================================="
                )

                HWR.beamline.detector.prepare_std_collection(
                    exp_time, nframes, filepath
                )

            self.log.debug("#COLLECT# Starting detector")

            # Check whether the live view monitoring is on. Restart if needed.
            process_name = os.getenv("MXCUBE_LIVEVIEW_NAME")
            command = [os.getenv("MXCUBE_LIVEVIEW")]
            if self.is_process_running(process_name) and self.is_process_running(
                "adxv"
            ):
                self.log.debug(f"{process_name} is already running.")
            else:
                os.system(f"killall -9 {process_name}")
                self.log.debug(f"{process_name} is not running. Starting...")
                self.start_process(command)

            if collection_type == "Characterization":
                self.log.debug("STARTING CHARACTERISATION")
                self.collect_characterisation(
                    start_angle, img_range, nframes, angle_inc, exp_time
                )

                # Move to 0 with default speed (fast)
                HWR.beamline.diffractometer.set_omega_velocity(self.default_speed)
                time.sleep(0.1)
                HWR.beamline.diffractometer.move_omega(0)

                # Adding H5 info for characterization
                start_angles_collected = []
                for nf in range(nframes):
                    start_angles_collected.append(start_angle + nf * angle_inc)
                self.add_h5_info_characterisation(
                    self.latest_h5_filename, start_angles_collected, img_range
                )

                latest_image = HWR.beamline.detector.get_eiger_name_pattern()
                latest_local_path = os.path.dirname(f"/gpfs{latest_image}_master.h5")
                latest_local_name = f"/gpfs{latest_image}_master.h5".split("/")[-1]
                self.write_info_txt(
                    latest_local_path,
                    latest_local_name,
                    start_angle,
                    nframes,
                    img_range,
                    angle_inc,
                    exp_time,
                    "screening",
                )

            else:
                self.log.debug("STARTING STANDARD COLLECTION")
                duration = self.total_angle_range / self.acq_speed
                start_time = time.time()

                latest_image = HWR.beamline.detector.get_eiger_name_pattern()
                latest_local_path = os.path.dirname(f"/gpfs{latest_image}_master.h5")
                latest_local_name = f"/gpfs{latest_image}_master.h5".split("/")[-1]

                # Start the progress emitter in a separate greenlet
                gevent.spawn(self.progress_emitter, start_time, duration)

                HWR.beamline.diffractometer.wait_omega()

                # Arm the detector here. For characterization it is a bit different.
                HWR.beamline.detector.start_acquisition()
                self.collect_std_collection(start_angle, stop_angle)

                # Move to 0 with default speed (fast)
                HWR.beamline.diffractometer.set_omega_velocity(self.default_speed)
                time.sleep(0.1)
                HWR.beamline.diffractometer.move_omega(0)

                self.add_h5_info_standard_data_collection(self.latest_h5_filename)

                self.write_info_txt(
                    latest_local_path,
                    latest_local_name,
                    start_angle,
                    nframes,
                    img_range,
                    stop_angle - start_angle,
                    exp_time,
                    "regular",
                )

        except RuntimeError:
            self.log.exception("")
        finally:
            self.acquisition_cleanup()

        latest_image = HWR.beamline.detector.get_eiger_name_pattern()
        latest_image = f"/gpfs{latest_image}_master.h5"
        self.adxv_notify(latest_image)

    def progress_emitter(self, start_time, duration):
        """Emits progress steps during data collection.

        Args:
            start_time (float): Start time of the data collection.
            duration (float): Estimated duration of the data collection.
        """
        while True:
            elapsed_time = time.time() - start_time
            progress = (elapsed_time / duration) * 100

            if progress >= 100:
                progress = 98
                self.emit("progressStep", progress)
                break
            self.emit("progressStep", progress)
            gevent.sleep(1)  # Non-blocking sleep

    def add_h5_info_standard_data_collection(self, imagepath):
        """Adds metadata to the HDF5 file for a standard data collection.

        Args:
            imagepath (str): Path to the HDF5 file.
        """
        time.sleep(1)
        self.log.debug("adding h5 info for standard data collection")
        try:
            f = h5py.File(imagepath, "r+")
            # source and instrument
            g = f.create_group("entry/source")
            g.attrs["NX_class"] = np.array("NXsource", dtype="S")
            g.create_dataset("name", data=np.array("PETRA III, DESY", dtype="S"))
            g = f.get("entry/instrument")
            g.create_dataset("name", data=np.array("P11", dtype="S"))
            # attenuator
            g = f.create_group("entry/instrument/attenuator")
            g.attrs["NX_class"] = np.array("NXattenuator", dtype="S")
            ds = g.create_dataset(
                "thickness", dtype="f8", data=float(self.get_filter_thickness())
            )
            ds.attrs["units"] = np.array("m", dtype="S")
            ds = g.create_dataset("type", data=np.array("Aluminum", dtype="S"))
            ds = g.create_dataset(
                "attenuator_transmission",
                dtype="f8",
                data=float(self.get_filter_transmission()),
            )

            # Keep it here as it is not clear if it is needed.
            # It was used in CC to fix the issue with the data processing

            # #fix rotation axis and detector orientation
            # ds = f.get(u"entry/sample/transformations/omega")
            # ds.attrs[u"vector"] = [1., 0., 0.]
            # ds = f.get(u"entry/instrument/detector/module/fast_pixel_direction")
            # ds.attrs[u"vector"] = [1., 0., 0.]
            # ds = f.get(u"entry/instrument/detector/module/slow_pixel_direction")
            # ds.attrs[u"vector"] = [0., 1., 0.]
            # delete phi angle info to avoid confusion
            nodes = [
                "entry/sample/goniometer/phi",
                "entry/sample/goniometer/phi_end",
                "entry/sample/goniometer/phi_range_average",
                "entry/sample/goniometer/phi_range_total",
            ]
            for node in nodes:
                if node in f:
                    del f[node]
            f.close()
        except RuntimeWarning:
            self.log.debug("writing header to H5 FAILED!")

    def add_h5_info_characterisation(
        self, imagepath, start_angles_collected, degreesperframe
    ):
        """Adds metadata to the HDF5 file for characterization data collection.

        Args:
            imagepath (str): Path to the HDF5 file.
            start_angles_collected (list): List of angles at which images were collected.
            degreesperframe (float): Number of degrees per frame collected.
        """
        time.sleep(1)
        self.log.debug("adding h5 info for characterization")
        try:
            f = h5py.File(imagepath, "r+")
            # source and instrument
            g = f.create_group("entry/source")
            g.attrs["NX_class"] = np.array("NXsource", dtype="S")
            g.create_dataset("name", data=np.array("PETRA III, DESY", dtype="S"))
            g = f.get("entry/instrument")
            g.create_dataset("name", data=np.array("P11", dtype="S"))
            # attenuator
            g = f.create_group("entry/instrument/attenuator")
            g.attrs["NX_class"] = np.array("NXattenuator", dtype="S")
            ds = g.create_dataset(
                "thickness", dtype="f8", data=float(self.get_filter_thickness())
            )
            ds.attrs["units"] = np.array("m", dtype="S")
            ds = g.create_dataset("type", data=np.array("Aluminum", dtype="S"))
            ds = g.create_dataset(
                "attenuator_transmission",
                dtype="f8",
                data=float(self.get_filter_transmission()),
            )
            # delete existing angle info
            nodes = [
                "entry/sample/goniometer/omega",
                "entry/sample/goniometer/omega_end",
                "entry/sample/goniometer/omega_range_average",
                "entry/sample/goniometer/omega_range_total",
                "entry/sample/goniometer/phi",
                "entry/sample/goniometer/phi_end",
                "entry/sample/goniometer/phi_range_average",
                "entry/sample/goniometer/phi_range_total",
                "entry/sample/transformations/omega",
                "entry/sample/transformations/omega_end",
                "entry/sample/transformations/omega_range_average",
                "entry/sample/transformations/omega_range_total",
            ]
            for node in nodes:
                if node in f:
                    del f[node]

            angles = []
            angles_end = []
            for angle in start_angles_collected:
                angles.append(float(angle))
                angles_end.append(float(angle + degreesperframe))
            g = f.get("entry/sample/goniometer")
            o = g.create_dataset("omega", dtype="f8", data=angles)
            o.attrs["vector"] = [1.0, 0.0, 0.0]
            g.create_dataset("omega_end", dtype="f8", data=angles_end)
            g.create_dataset(
                "omega_range_average", dtype="f8", data=float(degreesperframe)
            )
            g = f.get("entry/sample/transformations")
            o = g.create_dataset("omega", dtype="f8", data=angles)
            o.attrs["vector"] = [1.0, 0.0, 0.0]
            g.create_dataset("omega_end", dtype="f8", data=angles_end)
            g.create_dataset(
                "omega_range_average", dtype="f8", data=float(degreesperframe)
            )
            f.close()
        except RuntimeWarning:
            self.log.debug("writing header to H5 characterization FAILED!")

    def write_info_txt(
        self,
        path,
        name,
        startangle,
        frames,
        degreesperframe,
        imageinterval,
        exposuretime,
        run_type,
    ):
        """Writes info about the data collection into a text file.

        Args:
            path (str): Directory path where the info file will be saved.
            name (str): Name of the run.
            startangle (float): Starting angle of the collection.
            frames (int): Number of frames collected.
            degreesperframe (float): Degrees per frame.
            imageinterval (float): Interval between images.
            exposuretime (float): Exposure time in milliseconds.
            run_type (str): Type of run (e.g., 'regular' or 'screening').
        """
        if run_type == "regular":
            INFO_TXT = (
                "run type:            {run_type:s}\n"
                + "run name:            {name:s}\n"
                + "start angle:         {startangle:.2f}deg\n"
                + "frames:              {frames:d}\n"
                + "degrees/frame:       {degreesperframe:.2f}deg\n"
                + "exposure time:       {exposuretime:.3f}ms\n"
                + "energy:              {energy:.3f}keV\n"
                + "wavelength:          {wavelength:.3f}A\n"
                + "detector distance:   {detectordistance:.2f}mm\n"
                + "resolution:          {resolution:.2f}A\n"
                + "aperture:            {pinholeDiameter:d}um\n"
                + "focus:               {focus:s}\n"
                + "filter transmission: {filterTransmission:.3f}%\n"
                + "filter thickness:    {filterThickness:d}um\n"
                + "ring current:        {beamCurrent:.3f}mA\n"
                + "\n"
                + "For exact flux reading, please consult the staff."
            )

            energy = HWR.beamline.energy.get_value()
            wavelength = 12.3984 / (energy)  # in Angstrom
            resolution = HWR.beamline.resolution.get_value()
            detectordistance = HWR.beamline.detector.get_eiger_detector_distance()
            transmission = HWR.beamline.transmission.get_value()
            filter_thickness = self.get_filter_thickness_in_mm()
            pinhole_diameter = HWR.beamline.beam.get_pinhole_size()
            focus = HWR.beamline.beam.get_beam_focus_label()
            current = HWR.beamline.machine_info.get_current()

            output = INFO_TXT.format(
                run_type=run_type,
                name=name,
                startangle=startangle,
                frames=frames,
                degreesperframe=degreesperframe,
                exposuretime=exposuretime * 1000,
                energy=energy,
                wavelength=wavelength,
                detectordistance=detectordistance,
                resolution=resolution,
                pinholeDiameter=int(pinhole_diameter),
                focus=str(focus),
                filterTransmission=int(transmission),
                filterThickness=int(filter_thickness),
                beamCurrent=float(current),
            )

            f = open(path + "/info.txt", "w")
            f.write(output)
            f.close()

        if run_type == "screening":
            INFO_TXT = (
                "run type:            {run_type:s}\n"
                + "run name:            {name:s}\n"
                + "start angle:         {startangle:.2f}deg\n"
                + "frames:              {frames:d}\n"
                + "degrees/frame:       {degreesperframe:.2f}deg\n"
                + "exposure time:       {exposuretime:.3f}ms\n"
                + "energy:              {energy:.3f}keV\n"
                + "wavelength:          {wavelength:.3f}A\n"
                + "detector distance:   {detectordistance:.2f}mm\n"
                + "resolution:          {resolution:.2f}A\n"
                + "aperture:            {pinholeDiameter:d}um\n"
                + "focus:               {focus:s}\n"
                + "filter transmission: {filterTransmission:.3f}%\n"
                + "filter thickness:    {filterThickness:d}um\n"
                + "ring current:        {beamCurrent:.3f}mA\n"
                + "\n"
                + "For exact flux reading, please consult the staff."
            )

            energy = HWR.beamline.energy.get_value()
            wavelength = 12.3984 / (energy)  # in Angstrom
            resolution = HWR.beamline.resolution.get_value()
            detectordistance = HWR.beamline.detector.get_eiger_detector_distance()
            transmission = HWR.beamline.transmission.get_value()
            filter_thickness = self.get_filter_thickness_in_mm()
            pinhole_diameter = HWR.beamline.beam.get_pinhole_size()
            focus = HWR.beamline.beam.get_beam_focus_label()
            current = HWR.beamline.machine_info.get_current()

            output = INFO_TXT.format(
                run_type=run_type,
                name=name,
                startangle=startangle,
                frames=frames,
                degreesperframe=degreesperframe,
                exposuretime=exposuretime * 1000,
                energy=energy,
                wavelength=wavelength,
                detectordistance=detectordistance,
                resolution=resolution,
                pinholeDiameter=int(pinhole_diameter),
                focus=str(focus),
                filterTransmission=int(transmission),
                filterThickness=int(filter_thickness),
                beamCurrent=float(current),
            )

            f = open(path + "/info.txt", "w")
            f.write(output)
            f.close()

    def prepare_characterization(self):
        """Prepares for characterization data collection by setting the start angle and angle increment for the detector."""
        self.log.debug("Preparing for characterization data collection.")

        osc_pars = self.current_dc_parameters["oscillation_sequence"][0]
        start_angle = osc_pars["start"]
        HWR.beamline.detector.set_eiger_start_angle(start_angle)

        img_range = osc_pars["range"]
        HWR.beamline.detector.set_eiger_angle_increment(img_range)

    def collect_characterisation(
        self, start_angle, img_range, nimages, angle_inc, exp_time
    ):
        """Collects a series of images at different angles for characterization.

        Args:
            start_angle (float): Starting angle for the characterization acquisition.
            img_range (float): Range of angles over which a single image is collected.
            nimages (int): Number of images to be collected during the characterization process.
            angle_inc (float): Increment in angle between each image in the collection.
            exp_time (float): Exposure time for each image.
        """
        self.log.debug(
            "#COLLECT# Running OMEGA through the characterization acquisition"
        )

        start_angle = int(start_angle)

        for img_no in range(nimages):
            self.log.debug("collecting image %s" % img_no)
            start_at = start_angle + angle_inc * img_no
            stop_angle = start_at + img_range * 1.0

            self.log.debug("collecting image %s, angle %f" % (img_no, start_at))

            # [WIP]
            # NB! Another attempt to fix the misfires.
            # Keep comments here until finished
            # Here is the previous implementation:
            # self.collect_std_collection(start_at, stop_angle)

            # Here is slightly modified standard data collection routine
            # Adjust the angle since each time we are starting with 90 degrees offset.
            start_pos = start_at - self.turnback_time * self.acq_speed
            stop_pos = stop_angle + self.turnback_time * self.acq_speed

            self.log.debug("#COLLECT# Running OMEGA through the std acquisition")
            HWR.beamline.diffractometer.wait_omega_on()
            HWR.beamline.diffractometer.set_omega_velocity(self.default_speed)
            HWR.beamline.diffractometer.move_omega(start_pos)
            HWR.beamline.diffractometer.wait_omega_on()
            HWR.beamline.diffractometer.set_pso_control_arm(start_angle, stop_angle)
            HWR.beamline.diffractometer.set_omega_velocity(self.acq_speed)

            # Arm the detector only once in the beginning. Set to wait 4 triggers.
            if img_no == 0:
                HWR.beamline.detector.start_acquisition()
                time.sleep(3)
            HWR.beamline.diffractometer.move_omega(stop_pos)
            HWR.beamline.diffractometer.wait_omega_on()
            HWR.beamline.diffractometer.stop_motion()

            self.emit("progressStep", int(120 / (nimages) * (img_no + 1)))

    def prepare_std_collection(self, start_angle, img_range):
        """Prepares a standard collection by setting the start angle and angle increment in the detector's header.

        Args:
            start_angle (float): Starting angle for the standard collection sequence.
            img_range (float): Angle increment for each frame.

        Returns:
            bool: True if successful, False otherwise.
        """
        osc_pars = self.current_dc_parameters["oscillation_sequence"][0]
        start_angle = osc_pars["start"]
        HWR.beamline.detector.set_eiger_start_angle(start_angle)

        img_range = osc_pars["range"]
        HWR.beamline.detector.set_eiger_angle_increment(img_range)

    def collect_std_collection(self, start_angle, stop_angle):
        """Performs the standard data collection by moving the omega motor from start to stop angle.

        Args:
            start_angle (float): Starting angle for the collection.
            stop_angle (float): Stop angle for the collection.
        """
        start_pos = start_angle - self.turnback_time * self.acq_speed
        stop_pos = stop_angle + self.turnback_time * self.acq_speed

        self.log.debug("#COLLECT# Running OMEGA through the std acquisition")
        HWR.beamline.diffractometer.wait_omega_on()
        HWR.beamline.diffractometer.set_omega_velocity(self.default_speed)
        time.sleep(1)
        HWR.beamline.diffractometer.move_omega(start_pos)
        time.sleep(1)
        HWR.beamline.diffractometer.wait_omega_on()
        time.sleep(1)
        HWR.beamline.diffractometer.set_pso_control_arm(start_angle, stop_angle)
        time.sleep(3)
        HWR.beamline.diffractometer.set_omega_velocity(self.acq_speed)
        time.sleep(1)
        HWR.beamline.diffractometer.move_omega(stop_pos)

    def adxv_notify(self, image_filename, image_num=1):
        """Sends a notification to an ADXV to load an image file and display a specific slab.

        Args:
            image_filename (str): Filename of the image to be loaded.
            image_num (int, optional): Image number to load in ADXV. Defaults to 1.
        """
        self.log.info(f"ADXV notify {image_filename}")
        self.log.info(f"ADXV notify {image_num}")
        adxv_host = "localhost"
        adxv_port = 8100

        try:
            adxv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            adxv_socket.connect((adxv_host, adxv_port))
            adxv_socket.sendall(
                f"load_image {image_filename}\n slab {image_num}\n".encode()
            )
            adxv_socket.close()
        except RuntimeWarning:
            self.log.exception("")
        else:
            pass

    def is_process_running(self, process_name):
        """Checks if a process is running by its name.

        Args:
            process_name (str): Name of the process to check.

        Returns:
            bool: True if the process is running, False otherwise.
        """
        for proc in psutil.process_iter():
            if proc.name() == process_name:
                return True
        return False

    def start_process(self, command):
        """Starts a process with the specified command.

        Args:
            command (list): List of command arguments to start the process.
        """
        subprocess.Popen(
            command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            close_fds=True,
        )

    def acquisition_cleanup(self):
        """Performs cleanup after data acquisition, including stopping the detector and resetting motor velocities."""
        try:
            HWR.beamline.detector.stop_acquisition()
            HWR.beamline.diffractometer.wait_omega()
            HWR.beamline.diffractometer.set_omega_velocity(self.default_speed)
            self.log.debug("#COLLECT# Closing detector cover")
            HWR.beamline.diffractometer.detector_cover_close(wait=True)

            HWR.beamline.diffractometer.stop_motion()

        except RuntimeError:
            self.log.exception("")

    def get_or_create_group(self, parent_group, group_name):
        """Gets or creates a group within a parent group.

        Args:
            parent_group (h5py.Group): The parent group where the new group will be created.
            group_name (str): The name of the group to get or create.

        Returns:
            h5py.Group: The group object.
        """
        if group_name in parent_group:
            return parent_group[group_name]
        else:
            return parent_group.create_group(group_name)

    def create_or_get_dataset(self, group, dataset_name, dataset_data):
        """Creates or retrieves a dataset within a group.

        Args:
            group (h5py.Group): The group where the dataset will be created or retrieved.
            dataset_name (str): The name of the dataset.
            dataset_data: The data to be stored in the dataset.
        """
        if dataset_name in group:
            dataset = group[dataset_name]
        else:
            dataset = group.create_dataset(dataset_name, data=dataset_data)

    def get_filter_thickness(self):
        """Calculates the total thickness of three filters.

        Returns:
            float: The total thickness of the filters in meters. If the filter server is not available, it returns -1.
        """
        if self.filter_server:
            thick1 = self.filter_server.Filter1Thickness
            thick2 = self.filter_server.Filter2Thickness
            thick3 = self.filter_server.Filter3Thickness

            thickness = int(thick1) + int(thick2) + int(thick3)
            return float(thickness) / 1_000_000
        else:
            return -1

    def get_filter_thickness_in_mm(self):
        """Calculates the total thickness of three filters in millimeters.

        Returns:
            int: The total thickness of the filters in millimeters. If the filter server is not available, it returns -1.
        """
        if self.filter_server:
            thick1 = self.filter_server.Filter1Thickness
            thick2 = self.filter_server.Filter2Thickness
            thick3 = self.filter_server.Filter3Thickness

            thickness = int(thick1) + int(thick2) + int(thick3)

            return int(thickness)
        else:
            return -1

    def get_filter_transmission(self):
        """Gets the current transmission value from the filter server.

        Returns:
            float: The current transmission value. If the filter server is not available, it returns -1.
        """
        if self.filter_server:
            return self.filter_server.CurrentTransmission
        else:
            return -1

    def xdsapp_maxwell(self):
        """Starts XDSAPP auto-processing on the Maxwell cluster."""

        if HWR.beamline.session.get_beamtime_metadata() != None:
            self.log.debug("==== XDSAPP AUTOPROCESSING IS STARTED ==========")

            resolution = self.get_resolution()

            image_dir_local, filename = os.path.split(self.latest_h5_filename)

            image_dir = image_dir_local.replace(
                "/gpfs/current", HWR.beamline.session.get_beamtime_metadata()[2]
            )
            process_dir = image_dir.replace("/raw/", "/processed/")
            process_dir_local = image_dir_local.replace("/raw/", "/processed/")
            xdsapp_path = os.path.join(process_dir, "xdsapp")
            xdsapp_path_local = os.path.join(process_dir_local, "xdsapp")
            self.log.debug('============XDSAPP======== xdsapp_path="%s"' % xdsapp_path)
            self.log.debug(
                '============XDSAPP======== xdsapp_path_local="%s"' % xdsapp_path_local
            )

            try:
                self.mkdir_with_mode(xdsapp_path_local, mode=0o777)
                self.log.debug(
                    "=========== XDSAPP ============ XDSAPP directory created"
                )

            except OSError:
                self.log.debug(sys.exc_info())
                self.log.debug("Cannot create XDSAPP directory")

            base_process_dir = self.base_dir(process_dir_local, "processed")
            datasets_file = os.path.join(base_process_dir, "datasets.txt")

            try:
                open(datasets_file, "a", encoding="utf-8").write(
                    xdsapp_path_local.split("/gpfs/current/processed/")[1] + "\n"
                )
            except RuntimeError as err_msg:
                self.log.debug("Cannot write to datasets.txt")
                self.log.debug(sys.exc_info())

            ssh = HWR.beamline.session.get_ssh_command()
            sbatch = HWR.beamline.session.get_sbatch_command(
                jobname_prefix="xdsapp",
                logfile_path=xdsapp_path.replace(
                    HWR.beamline.session.get_beamtime_metadata()[2],
                    "/beamline/p11/current",
                )
                + "/xdsapp.log",
            )

            cmd = (
                "/asap3/petra3/gpfs/common/p11/processing/xdsapp_sbatch.sh "
                + "{imagepath:s} {processpath:s} {res:f}"
            ).format(
                imagepath=image_dir + "/" + filename,
                processpath=xdsapp_path.replace(
                    HWR.beamline.session.get_beamtime_metadata()[2],
                    "/beamline/p11/current",
                ),
                res=resolution,
            )

            os.system(
                '{ssh:s} "{sbatch:s} --wrap \\"{cmd:s}\\""'.format(
                    ssh=ssh, sbatch=sbatch, cmd=cmd
                )
            )
        else:
            self.log.debug(
                "Beamtime metadata is not found. No online resouces available. Autoprocessing on the cluster is not available."
            )

    def autoproc_maxwell(self):
        """Starts AutoProc auto-processing on the Maxwell cluster."""

        if HWR.beamline.session.get_beamtime_metadata() != None:
            self.log.debug("==== AUTOPROC AUTOPROCESSING IS STARTED ==========")

            resolution = self.get_resolution()

            image_dir_local, filename = os.path.split(self.latest_h5_filename)

            image_dir = image_dir_local.replace(
                "/gpfs/current", HWR.beamline.session.get_beamtime_metadata()[2]
            )
            process_dir = image_dir.replace("/raw/", "/processed/")
            process_dir_local = image_dir_local.replace("/raw/", "/processed/")
            autoproc_path = os.path.join(process_dir, "autoproc")
            autoproc_path_local = os.path.join(process_dir_local, "autoproc")
            self.log.debug(
                '============AUTOPROC======== autoproc_path="%s"' % autoproc_path
            )
            self.log.debug(
                '============AUTOPROC======== autoproc_path_local="%s"'
                % autoproc_path_local
            )

            try:
                self.mkdir_with_mode(autoproc_path_local, mode=0o777)
                self.log.debug(
                    "=========== AUTOPROC ============ autoproc directory created"
                )

            except OSError:
                self.log.debug(sys.exc_info())
                self.log.debug("Cannot create AUTOPROC directory")

            base_process_dir = self.base_dir(process_dir_local, "processed")
            datasets_file = os.path.join(base_process_dir, "datasets.txt")

            try:
                open(datasets_file, "a", encoding="utf-8").write(
                    autoproc_path_local.split("/gpfs/current/processed/")[1] + "\n"
                )
            except RuntimeError as err_msg:
                self.log.debug("Cannot write to datasets.txt")
                self.log.debug(sys.exc_info())

            ssh = HWR.beamline.session.get_ssh_command()
            sbatch = HWR.beamline.session.get_sbatch_command(
                jobname_prefix="autoproc",
                logfile_path=autoproc_path.replace(
                    HWR.beamline.session.get_beamtime_metadata()[2],
                    "/beamline/p11/current",
                )
                + "/autoproc.log",
            )

            cmd = (
                "/asap3/petra3/gpfs/common/p11/processing/autoproc_sbatch.sh "
                + "{imagepath:s} {processpath:s}"
            ).format(
                imagepath=image_dir + "/" + filename,
                processpath=autoproc_path.replace(
                    HWR.beamline.session.get_beamtime_metadata()[2],
                    "/beamline/p11/current",
                )
                + "/processing",
            )

            os.system(
                '{ssh:s} "{sbatch:s} --wrap \\"{cmd:s}\\""'.format(
                    ssh=ssh, sbatch=sbatch, cmd=cmd
                )
            )
        else:
            self.log.debug(
                "Beamtime metadata is not found. No online resouces available. Autoprocessing on the cluster is not available."
            )

    def trigger_auto_processing(self, process_event=None, frame_number=None):
        """Triggers auto processing based on the experiment type.

        Args:
            process_event: Optional event that triggered the auto processing.
            frame_number (int, optional): Number of frames to process.
        """
        self.log.debug("Triggering auto processing")
        collection_type = self.current_dc_parameters["experiment_type"]
        self.log.debug(
            "=============== Supported experiment types: ===========\n"
            + str(self.current_dc_parameters["experiment_type"])
        )

        self.xdsapp_maxwell()
        self.autoproc_maxwell()

    def diffractometer_prepare_collection(self):
        """Prepares the diffractometer for data collection.

        Returns:
            bool: True if the diffractometer is in collect phase, False otherwise.
        """
        self.log.debug("#COLLECT# preparing collection ")
        if not HWR.beamline.diffractometer.is_collect_phase():
            self.log.debug("#COLLECT# going to collect phase")
            HWR.beamline.diffractometer.goto_collect_phase(wait=True)

        self.log.debug(
            "#COLLECT# now in collect phase: %s"
            % HWR.beamline.diffractometer.is_collect_phase()
        )

        return HWR.beamline.diffractometer.is_collect_phase()

    def get_relative_path(self, path1, path2):
        """Returns the relative path from path1 to path2.

        Args:
            path1 (str): First path.
            path2 (str): Second path.

        Returns:
            str: The relative path between path1 and path2.
        """
        path_1 = path1.split(os.path.sep)
        path_2 = path2.split(os.path.sep)

        for i, v__ in enumerate(path_2):
            if path_1[i] != v__:
                break

            parts = [".."] * (len(path_2) - i)
            parts.extend(path_1[i:])

        return os.path.join(*parts)

    def base_dir(self, path, what):
        """Returns the base directory path that contains the specified "what" directory or file.

        Args:
            path (str): The path to a file or directory.
            what (str): The directory or file name to search for.

        Returns:
            str: The base directory containing the "what".
        """
        what = what.lstrip(os.path.sep)

        if path.startswith(os.path.sep):
            start_sep = os.path.sep
        else:
            start_sep = ""

        path_ = path.split(os.path.sep)
        for i, v__ in enumerate(path_):
            if path_[i] == what:
                break

        return start_sep + os.path.join(*path_[: i + 1])

    def mkdir_with_mode(self, directory, mode):
        """Creates a directory with the specified mode.

        Args:
            directory (str): Path of the directory to create.
            mode (int): Mode (permissions) to set for the directory.
        """
        if not os.path.isdir(directory):
            oldmask = os.umask(000)
            os.makedirs(directory, mode=mode)
            os.umask(oldmask)

            self.log.debug("local directory created")

    def create_directories(self, *args):
        """Creates directories for raw files and processing files."""
        for directory in args:
            try:
                self.mkdir_with_mode(directory, mode=0o777)
            except os.error as err_:
                if err_.errno != errno.EEXIST:
                    raise

    def check_path(self, path=None, force=False):
        """Checks if a path is valid and accessible, and creates directories if needed.

        Args:
            path (str, optional): The path to check. Defaults to None.
            force (bool, optional): Whether to create the directories if they don't exist. Defaults to False.

        Returns:
            str or bool: The path if valid and accessible, or False if not.
        """
        path = str(path).replace("\\", "/")
        dirs = path.strip("/").split("/")
        path = ""
        for dir_ in dirs:
            if dir_.find("*") > -1 or dir_.find("?") > -1 or dir_.find("..") > -1:
                return False
            path += "/" + dir_
            if not os.access(path, os.F_OK):
                if force:
                    try:
                        os.mkdir(path, mode=0o777)
                    except RuntimeError:
                        self.log.debug("mkdir failed:", str(sys.exc_info()))

                        self.log.exception("")
                        return False
                else:
                    self.log.debug("dir not found:", str(sys.exc_info()))
                    return False
        if not os.access(path, os.W_OK):
            self.log.debug("dir not writable:", str(sys.exc_info()))
            return False
        return path

    def create_characterisation_directories(self):
        """Creates directories for raw files and processing files for EDNA and MOSFLM."""
        self.create_directories(
            self.current_dc_parameters["fileinfo"]["directory"],
            self.current_dc_parameters["fileinfo"]["process_directory"],
        )

        collection_type = HWR.beamline.collect.current_dc_parameters["experiment_type"]
        self.log.debug("PREPARING FOLDERS FOR COLLECTION TYPE %s", str(collection_type))

        xds_directory, auto_directory = self.prepare_input_files()
        xds_directory = xds_directory.replace("/rotational_", "/screening_").replace(
            "/xdsapp", "/edna"
        )
        auto_directory = auto_directory.replace("/rotational_", "/screening_").replace(
            "/xdsapp", "/edna"
        )
        try:
            self.create_directories(xds_directory, auto_directory)
            os.system("chmod -R 777 %s %s" % (xds_directory, auto_directory))
        except Exception:
            logging.exception("Could not create processing file directory")
            return
        if xds_directory:
            self.current_dc_parameters["xds_dir"] = xds_directory
        if auto_directory:
            self.current_dc_parameters["auto_dir"] = auto_directory

    def create_file_directories(self):
        """Creates directories for raw files and processing files."""
        self.create_directories(
            self.current_dc_parameters["fileinfo"]["directory"],
            self.current_dc_parameters["fileinfo"]["process_directory"],
        )

        xds_directory, auto_directory = self.prepare_input_files()
        try:
            self.create_directories(xds_directory, auto_directory)
            os.system("chmod -R 777 %s %s" % (xds_directory, auto_directory))
        except Exception:
            logging.exception("Could not create processing file directory")
            return
        if xds_directory:
            self.current_dc_parameters["xds_dir"] = xds_directory
        if auto_directory:
            self.current_dc_parameters["auto_dir"] = auto_directory

    def prepare_input_files(self):
        """Prepares directories for processing input files.

        Returns:
            tuple: Paths for XDS and AutoProc directories.
        """
        logging.getLogger("user_level_log").info(
            "Creating XDS processing input file directories"
        )

        xds_input_file_dirname = (
            "%s" % (self.current_dc_parameters["fileinfo"]["prefix"],)
            + "/rotational_"
            + str(self.current_dc_parameters["fileinfo"]["run_number"]).zfill(3)
        )
        xds_directory = os.path.join(
            self.current_dc_parameters["fileinfo"]["directory"].replace(
                "/current/raw", "/current/processed"
            ),
            xds_input_file_dirname,
            "xdsapp",
        )

        auto_directory = xds_directory

        self.log.info(
            "[COLLECT] Processing input file directories: XDS: %s, AUTO: %s"
            % (xds_directory, auto_directory)
        )
        return xds_directory, auto_directory

    def take_crystal_snapshots(self):
        """Takes sample snapshots and saves them to disk."""
        if self.current_dc_parameters["take_snapshots"]:
            snapshot_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["directory"], "snapshot"
            )
            if not os.path.exists(snapshot_directory):
                try:
                    self.create_directories(snapshot_directory)
                except Exception:
                    self.log.exception("Collection: Error creating snapshot directory")

            number_of_snapshots = self.current_dc_parameters["take_snapshots"]
            logging.getLogger("user_level_log").info(
                "Collection: Taking %d sample snapshot(s)" % number_of_snapshots
            )
            if HWR.beamline.diffractometer.get_current_phase() != "Centring":
                logging.getLogger("user_level_log").info(
                    "Moving Diffractometer to CentringPhase"
                )
                HWR.beamline.diffractometer.goto_centring_phase(wait=True)
                self.move_to_centered_position()

            for snapshot_index in range(number_of_snapshots):
                snapshot_filename = os.path.join(
                    snapshot_directory,
                    "%s_%s_%s.snapshot.jpeg"
                    % (
                        self.current_dc_parameters["fileinfo"]["prefix"],
                        self.current_dc_parameters["fileinfo"]["run_number"],
                        (snapshot_index + 1),
                    ),
                )
                self.current_dc_parameters[
                    "xtalSnapshotFullPath%i" % (snapshot_index + 1)
                ] = snapshot_filename
                self._take_crystal_snapshot(snapshot_filename)
                time.sleep(1)  # needed, otherwise will get the same images
                if number_of_snapshots > 1:
                    HWR.beamline.diffractometer.move_omega_relative(90)
                    time.sleep(1)  # needed, otherwise will get the same images



import json
import logging
import shutil
from collections import defaultdict
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List, Optional
from zoneinfo import ZoneInfo

import requests
from pydantic import ValidationError
from pyicat_plus.client.main import IcatClient

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractLims import AbstractLims
from mxcubecore.model.lims_session import (
    Download,
    Lims,
    LimsSessionManager,
    SampleInformation,
    SampleSheet,
    Session,
)

logger = logging.getLogger("HWR")


class ICATLIMS(AbstractLims):
    """
    ICAT+ client.
    """

    def __init__(self, name):
        super().__init__(name)
        HardwareObject.__init__(self, name)
        self.investigations = None
        self.icatClient = None
        self.lims_rest = None
        self.ingesters = None

    def init(self):
        self.url = self.get_property("ws_root")
        self.ingesters = self.get_property("queue_urls")
        self.investigations = []
        self.samples = []

        # Initialize ICAT client
        self.icatClient = IcatClient(
            icatplus_restricted_url=self.url,
            metadata_urls=["bcu-mq-01:61613"],
            reschedule_investigation_urls=["bcu-mq-01:61613"],
        )

    def get_lims_name(self) -> List[Lims]:
        return [
            Lims(name="DRAC", description="Data Repository for Advancing open sCience"),
        ]

    def _create_icat_session(self, user_name: str, password: str):
        self.icat_session: dict = self.icatClient.do_log_in(password)

    def login(
        self,
        user_name: str,
        password: str,
        session_manager: Optional[LimsSessionManager],
    ) -> LimsSessionManager:
        msg = f"authenticate {user_name}"
        logger.debug(msg)

        self._create_icat_session(user_name=user_name, password=password)

        if self.icatClient is None or self.icatClient is None:
            msg = "Error initializing icatClient: "
            msg += f"icatClient={self.url}"
            logger.error(msg)
            raise RuntimeError("Could not initialize icatClient")

        # Connected to metadata icatClient
        msg = "Connected succesfully to icatClient: "
        msg += f"fullName={self.icat_session['fullName']}, url={self.url}"
        logger.debug(msg)

        # Retrieving user's investigations
        sessions = self.to_sessions(self.__get_all_investigations())

        if len(sessions) == 0:
            msg = f"No sessions available for user {user_name}"
            raise RuntimeError(msg)

        msg = f"Successfully retrieved {len(sessions)} sessions"
        logger.debug(msg)

        # This is done because ICATLims can be used standalone or from ESRFLims
        if session_manager is not None:
            self.session_manager = session_manager

        # Check if there is currently a session in use and if user have
        # access to that session
        if self.session_manager.active_session:
            session_found = False
            session_id = self.session_manager.active_session.session_id
            for session in sessions:
                if session.session_id == session_id:
                    session_found = True
                    break

            if not session_found:
                msg = f"Current session in-use (with id {session_id}) "
                msg += f"not avaialble for user {user_name}"
                raise RuntimeError(msg)

        return self.session_manager, self.icat_session["name"], sessions

    def is_user_login_type(self) -> bool:
        return True

    def get_proposals_by_user(self, user_name):
        msg = f"get_proposals_by_user {user_name}\n"
        msg += f"[ICATCLient] Read {len(self.lims_rest.investigations)} investigations"
        logger.debug(msg)

        return self.lims_rest.to_sessions(self.lims_rest.investigations)

    def _get_loaded_pucks(self, parcels) -> list:
        """Retrieve all pucks from the parcels that have a defined
            'sampleChangerLocation'.
            A puck is considered "loaded" if it contains the key
            'sampleChangerLocation'.
            Iterates through all parcels and collects such pucks.
        Returns:
            A list of pucks (dicts) that have 'sampleChangerLocation' defined.
        """
        loaded_pucks = []

        if parcels:
            for parcel in parcels:
                pucks = parcel.get("content", [])
                for puck in pucks:
                    if "sampleChangerLocation" in puck:
                        # Add information about the parcel
                        puck["parcelName"] = parcel.get("name")
                        puck["parcelId"] = parcel.get("id")
                        loaded_pucks.append(puck)

        return loaded_pucks

    def get_samples(self, lims_name: str) -> list:
        """Retrieve and process sample information from LIMS based on the
            provided name:
            - Retrieves parcel data (containers like UniPucks or SpinePucks).
            - Retrieves sample sheet data.
            - Identifies and processes only loaded pucks
              (those with a 'sampleChangerLocation').
            - Converts each sample in the pucks into internal queue samples
              using `__to_sample`.
        Args:
            The LIMS name or identifier used to fetch sample-related data.

        Returns:
            A list of processed sample objects ready for queuing.
        """

        self.samples = []

        try:
            session = self.session_manager.active_session
            logger.debug(
                "[ICATClient] get_samples: session_id=%s, proposal_name=%s",
                session.session_id,
                session.proposal_name,
            )

            # Load parcels (pucks)
            self.parcels = self.get_parcels()

            # Load sample sheets
            self.sample_sheets = self.get_samples_sheets()
            logger.debug(
                "[ICATClient] %d sample sheets retrieved", len(self.sample_sheets)
            )

            # Filter for loaded pucks
            self.loaded_pucks = self._get_loaded_pucks(self.parcels)
            msg = f"[ICATClient] {len(self.loaded_pucks)} loaded pucks found"
            logger.debug(msg)

            # Extract and process samples from loaded pucks
            for puck in self.loaded_pucks:
                tracking_samples = puck.get("content", [])
                puck_name = puck.get("name", "Unnamed")
                location = puck.get("sampleChangerLocation", "Unknown")
                msg = f"[ICATClient] Found puck {puck_name} at position "
                msg += f"{location}, containing {len(tracking_samples)} samples"
                logger.debug(msg)
                for tracking_sample in tracking_samples:
                    sample = self.__to_sample(tracking_sample, puck, self.sample_sheets)
                    self.samples.append(sample)
        except RuntimeError:
            logger.exception("[ICATClient] Error retrieving samples: %s")
        else:
            msg = f"[ICATClient] Total {len(self.samples)} samples read"
            logger.debug(msg)
            return self.samples
        return []

    def get_sample_sheet_by_id(
        self, samples: List[SampleSheet], sample_id: int
    ) -> Optional[SampleSheet]:
        """
        Retrieve a sample sheet by its unique ID.

        Args:
            samples (List[SampleSheet]): A list of Sample objects.
            sample_id (int): The unique identifier of the sample sheet to retrieve.

        Returns:
            Optional[Sample]: The Sample object if found, otherwise None.
        """
        return next((sample for sample in samples if sample.id == sample_id), None)

    def objectid_to_int(self, oid_str):
        return int(oid_str, 16)

    def int_to_objectid(self, i):
        return hex(i)[2:].zfill(24)

    def __add_download_path_to_processing_plan(
        self, processing_plan, downloads: List[Download]
    ):
        file_path_lookup = {d.filename: d.path for d in downloads}
        group_paths = defaultdict(list)
        for d in downloads:
            if d.groupName is not None:
                group_paths[d.groupName].append(d.path)

        # Enrich the processing_plan
        for item in processing_plan:
            key = item.get("key")
            value = item.get("value")
            if (
                key == "reference"
                and isinstance(value, str)
                and value in file_path_lookup
            ):
                item["value"] = {"filepath": file_path_lookup[value]}
            if key == "search_models":
                models = value
                if isinstance(models, str):
                    models = json.loads(models)
                for model in models:
                    group = model.get("pdb_group")
                    if group in group_paths:
                        model["file_paths"] = group_paths[group]
                item["value"] = models

    def _safe_json_loads(self, json_str):
        try:
            return json.loads(json_str)
        except Exception:
            return str(json_str)

    def __extract_sample_identifiers(self, tracking_sample: dict, puck: dict) -> dict:
        # Basic identifiers
        sample_name = str(tracking_sample.get("name"))

        # MXCuBE needs to be an integer while in DRAC is a ObjectId
        # Mongo @BES needs to be smaller then 8 bytes
        sample_id = int(str(self.objectid_to_int(tracking_sample.get("id")))[-6:])
        # id to the sample sheet declared in the user portal
        sample_sheet_id = tracking_sample.get("sampleId")
        # identifier that points to the sample tracking
        tracking_sample_id = tracking_sample.get("_id")

        msg = f"[ICATClient] Sample ids sample_id={sample_id} "
        msg += f"sample_sheet_id={sample_sheet_id} "
        msg += f"tracking_sample_id={tracking_sample_id}"
        logger.debug(msg)

        sample_location = tracking_sample.get("sampleContainerPosition")
        puck_location = str(puck.get("sampleChangerLocation", "Unknown"))
        puck_name = puck.get("name", "UnknownPuck")
        parcel_name = puck.get("parcelName")
        parcel_id = puck.get("parcelId")

        protein_acronym = self.__resolve_protein_acronym(sample_name, sample_sheet_id)

        return {
            "sampleName": sample_name,
            "sampleId": sample_id,
            "sample_sheet_id": sample_sheet_id,
            "trackingSampleId": tracking_sample_id,
            "proteinAcronym": protein_acronym,
            "sampleLocation": sample_location,
            "containerCode": puck_name,
            "containerSampleChangerLocation": puck_location,
            "SampleTrackingParcel_name": parcel_name,
            "SampleTrackingParcel_id": parcel_id,
            "SampleTrackingContainer_id": puck_name,
            "SampleTrackingContainer_name": parcel_id,
        }

    def __resolve_protein_acronym(self, sample_name: str, sample_sheet_id: str) -> str:
        sample_sheet = self.get_sample_sheet_by_id(self.sample_sheets, sample_sheet_id)
        return sample_sheet.name if sample_sheet else sample_name

    def __parse_experiment_plan(self, tracking_sample: dict) -> dict[str, Any]:
        return {
            item["key"]: item["value"]
            for item in tracking_sample.get("experimentPlan", {})
        }

    def __build_diffraction_plan(self, experiment_plan: dict) -> dict[str, Any]:
        return {
            # "diffractionPlanId": 457980, TODO: do we need this?
            "experimentKind": experiment_plan.get("experimentKind"),
            "numberOfPositions": experiment_plan.get("numberOfPositions"),
            "observedResolution": experiment_plan.get("observedResolution"),
            "preferredBeamDiameter": experiment_plan.get("preferredBeamDiameter"),
            "radiationSensitivity": experiment_plan.get("radiationSensitivity"),
            "requiredCompleteness": experiment_plan.get("requiredCompleteness"),
            "requiredMultiplicity": experiment_plan.get("requiredMultiplicity"),
            "requiredResolution": experiment_plan.get("requiredResolution"),
        }

    def __prepare_processing_plan(
        self, tracking_sample: dict, sample_sheet_id: str, protein_acronym: str
    ) -> dict[str, Any]:
        processing_plan = tracking_sample.get("processingPlan", [])
        if not processing_plan:
            return {}

        # Convert string values to JSON if possible
        for item in processing_plan:
            item["value"] = self._safe_json_loads(item.get("value"))

        downloads = self.__get_or_download_plan_resources(
            sample_sheet_id, protein_acronym
        )

        if downloads:
            try:
                self.__add_download_path_to_processing_plan(processing_plan, downloads)
            except RuntimeError:
                logger.exception("Failed __add_download_path_to_processing_plan")
        return {item["key"]: item["value"] for item in processing_plan}

    def __get_or_download_plan_resources(
        self, sample_sheet_id: str, protein_acronym: str
    ) -> List[Download]:
        if not hasattr(self, "_downloads_cache"):
            self._downloads_cache = {}

        cache_key = (sample_sheet_id, protein_acronym)
        sample_information = self.__get_sample_information_by(sample_sheet_id)
        if not sample_information:
            return []

        cached = self._downloads_cache.get(cache_key)

        # Validate cache by comparing resource count
        if cached and len(cached) == len(sample_information.resources):
            logger.debug(f"Reusing cached downloads for {cache_key}")
            return cached

        # Otherwise, re-download
        # create subfolder per protein acronym
        destination_folder = (
            Path(HWR.beamline.session.get_base_process_directory())
            / "processing_plan_resources"
            / protein_acronym
        )
        destination_folder.mkdir(parents=True, exist_ok=True)

        logger.debug(
            f"Download resource: sample_sheet_id={sample_sheet_id} "
            f"destination_folder={destination_folder}"
        )

        downloads = self._download_resources(
            sample_sheet_id, sample_information.resources, destination_folder, ""
        )

        logger.debug(f"Downloaded {len(downloads)} resources")
        self._downloads_cache[cache_key] = downloads
        return downloads

    def __to_sample(
        self, tracking_sample: dict, puck: dict, sample_sheets: List[SampleSheet]
    ) -> dict[str, Any]:
        """
        Convert a tracking sample and associated metadata into the internal
        sample data structure.
        - Extracts relevant sample metadata.
        - Resolves protein acronym from the sample sheet if available.
        - Maps experiment plan details into a diffraction plan dictionary.
        - Assembles all relevant fields into a structured sample dictionary.

        Args:
            tracking_sample (dict): The raw sample data from tracking.
            puck (dict): The puck (container) metadata associated with the sample.
            sample_sheets (List[SampleSheet]): List of sample sheets used for lookup.

        Returns:
            dict: A dictionary representing the standardized internal sample format.
        """
        sample_id_info = self.__extract_sample_identifiers(tracking_sample, puck)
        experiment_plan = self.__parse_experiment_plan(tracking_sample)
        processing_plan = self.__prepare_processing_plan(
            tracking_sample,
            sample_id_info["sample_sheet_id"],
            sample_id_info["proteinAcronym"],
        )

        return {
            **sample_id_info,
            "experimentType": experiment_plan.get("workflowType"),
            "crystalSpaceGroup": experiment_plan.get("forceSpaceGroup"),
            "diffractionPlan": self.__build_diffraction_plan(experiment_plan),
            "experimentPlan": experiment_plan,
            "processingPlan": processing_plan,
            "comments": tracking_sample.get("comments"),
        }

    def create_session(self, session_dict):
        pass

    def _store_data_collection_group(self, group_data):
        pass

    @property
    def only_staff_session_selection(self) -> bool:
        return bool(
            self.get_property("only_staff_session_selection", default_value=False)
        )

    def store_robot_action(self, proposal_id: str):
        raise NotImplementedError

    @property
    def filter(self):
        return self.get_property("filter", None)

    @property
    def override_beamline_name(self):
        return self.get_property(
            "override_beamline_name", HWR.beamline.session.beamline_name
        )

    @property
    def compatible_beamlines(self):
        return self.get_property(
            "compatible_beamlines", HWR.beamline.session.beamline_name
        )

    @property
    def data_portal_url(self):
        return self.get_property("data_portal_url", None)

    @property
    def user_portal_url(self):
        return self.get_property("user_portal_url", None)

    @property
    def logbook_url(self):
        return self.get_property("logbook_url", None)

    @property
    def before_offset_days(self):
        return self.get_property("before_offset_days", "1")

    @property
    def after_offset_days(self):
        return self.get_property("after_offset_days", "1")

    def _string_to_format_date(self, date: str, fmt: str) -> str:
        if date is not None:
            date_time = self._tz_aware_fromisoformat(date)
            if date_time is not None:
                return date_time.strftime(fmt)
        return ""

    def _string_to_date(self, date: str) -> str:
        return self._string_to_format_date(date, "%Y%m%d")

    def _string_to_time(self, date: str) -> str:
        return self._string_to_format_date(date, "%H:%M:%S")

    def _tz_aware_fromisoformat(self, date: str) -> datetime:
        try:
            return datetime.fromisoformat(date).astimezone()
        except (TypeError, ValueError):
            return None

    def set_active_session_by_id(self, session_id: str) -> Session:
        logger.debug(f"set_active_session_by_id: {session_id}")

        if self.is_session_already_active(self.session_manager.active_session):
            return self.session_manager.active_session

        sessions = self.session_manager.sessions

        logger.debug(f"Sessions: {len(sessions)}")

        if len(sessions) == 0:
            logger.warning("Session list is empty. No session candidates")
            raise RuntimeError("No sessions available")

        if len(sessions) == 1:
            self.session_manager.active_session = sessions[0]
            logger.debug(
                "Session list contains a single session. proposal_name=%s",
                self.session_manager.active_session.proposal_name,
            )
            return self.session_manager.active_session

        session_list = [obj for obj in sessions if obj.session_id == session_id]
        if len(session_list) != 1:
            raise RuntimeError(
                "Session not found in the local list of sessions. session_id="
                + session_id
            )
        self.session_manager.active_session = session_list[0]
        return self.session_manager.active_session

    def allow_session(self, session: Session):
        self.active_session = session
        logger.debug("allow_session investigationId=%s", session.session_id)
        self.icatClient.reschedule_investigation(session.session_id)

    def get_session_by_id(self, sid: str):
        msg = f"get_session_by_id investigationId={sid} "
        msg += f"investigations={len(self.investigations)}"
        logger.debug(msg)

        investigation_list = list(filter(lambda p: p["id"] == sid, self.investigations))
        if len(investigation_list) == 1:
            self.investigation = investigation_list[0]
            return self.__to_session(investigation_list[0])

        logger.warning("No investigation found")
        return None

    def __get_all_investigations(self):
        """Returns all investigations by user. An investigation corresponds to
        one experimental session. It returns an empty array in case of error"""
        self.investigations = []
        try:
            msg = f"__get_all_investigations before={self.before_offset_days} "
            msg += f"after={self.after_offset_days} "
            msg += f"beamline={self.override_beamline_name} "
            msg += (
                f"isInstrumentScientist={self.icat_session['isInstrumentScientist']} "
            )
            msg += f"isAdministrator={self.icat_session['isAdministrator']} "
            msg += f"compatible_beamlines={self.compatible_beamlines}"
            logger.debug(msg)

            if self.icat_session is not None and (
                self.icat_session["isAdministrator"]
                or self.icat_session["isInstrumentScientist"]
            ):
                # Setting up of the session done by admin or staff
                self.investigations = self.icatClient.get_investigations_by(
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                    instrument_name=self.compatible_beamlines,
                )
            elif self.only_staff_session_selection:
                if self.session_manager.active_session is None:
                    # print warning an return no investigations
                    # if no session selected and only staff is allowed
                    logger.warning(
                        "No session selected. Only staff can select a session"
                    )
                    return []

                self.investigations = self.icatClient.get_investigations_by(
                    ids=[self.session_manager.active_session.session_id]
                )
            else:
                self.investigations = self.icatClient.get_investigations_by(
                    filter=self.filter,
                    instrument_name=self.compatible_beamlines,
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                )
        except Exception:
            self.investigations = []
            logger.exception("Failed on __get_all_investigations")
        else:
            msg = "__get_all_investigations retrieved "
            msg += f"{len(self.investigations)} investigations"
            logger.debug(msg)

        return self.investigations

    def __get_proposal_number_by_investigation(self, investigation):
        """
        Given an investigation it returns the proposal number.
        Example: investigation["name"] = "MX-1234"
        returns: 1234

        TODO: this might not work for all type of proposals (example: TEST proposals)
        """
        return (
            investigation["name"]
            .replace(investigation["type"]["name"], "")
            .replace("-", "")
        )

    def _get_data_portal_url(self, investigation):
        try:
            return (
                self.data_portal_url.replace("{id}", str(investigation["id"]))
                if self.data_portal_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_logbook_url(self, investigation):
        try:
            return (
                self.logbook_url.replace("{id}", str(investigation["id"]))
                if self.logbook_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_user_portal_url(self, investigation):
        try:
            return (
                self.user_portal_url.replace(
                    "{id}", str(investigation["parameters"]["Id"])
                )
                if self.user_portal_url is not None
                and investigation["parameters"]["Id"] is not None
                else ""
            )
        except Exception:
            return ""

    def __get_investigation_parameter_by_name(
        self, investigation: dict, parameter_name: str
    ) -> str:
        """
        Gets the metadata of the parameters in an investigation
        Returns the value of the specified parameter if it exists,
        otherwise returns an empty string.
        """
        return investigation.get("parameters", {}).get(parameter_name, None)

    def __to_session(self, investigation) -> Session:
        """This methods converts a ICAT investigation into a session"""

        actual_start_date = (
            investigation["parameters"]["__actualStartDate"]
            if "__actualStartDate" in investigation["parameters"]
            else investigation["startDate"]
        )
        actual_end_date = (
            investigation["parameters"]["__actualEndDate"]
            if "__actualEndDate" in investigation["parameters"]
            else investigation.get("endDate", None)
        )

        instrument_name = investigation["instrument"]["name"]

        # If session has been rescheduled new date is overwritten
        return Session(
            code=investigation["type"]["name"],
            number=self.__get_proposal_number_by_investigation(investigation),
            title=f"{investigation['title']}",
            session_id=str(investigation["id"]),
            proposal_id=str(investigation["id"]),
            proposal_name=investigation["name"],
            beamline_name=instrument_name,
            comments="",
            start_datetime=investigation.get(
                "startDate", None
            ),  # self._string_to_date(investigation.get("startDate", None)),
            start_date=self._string_to_date(investigation.get("startDate", None)),
            start_time=self._string_to_time(investigation.get("startDate", None)),
            end_datetime=investigation.get("endDate", None),
            end_date=self._string_to_date(
                investigation.get("endDate", None)
            ),  # self._string_to_time(investigation.get("endDate", None)),
            end_time=self._string_to_time(investigation.get("endDate", None)),
            actual_start_date=self._string_to_date(actual_start_date),
            actual_start_time=self._string_to_time(actual_start_date),
            actual_end_date=self._string_to_date(actual_end_date),
            actual_end_time=self._string_to_time(actual_end_date),
            nb_shifts="3",
            scheduled=str(self.is_scheduled_on_host_beamline(instrument_name)),
            is_scheduled_time=self.is_scheduled_now(actual_start_date, actual_end_date),
            is_scheduled_beamline=self.is_scheduled_on_host_beamline(instrument_name),
            data_portal_URL=self._get_data_portal_url(investigation),
            user_portal_URL=self._get_user_portal_url(investigation),
            logbook_URL=self._get_logbook_url(investigation),
            is_rescheduled=bool("__actualEndDate" in investigation["parameters"]),
            volume=self.__get_investigation_parameter_by_name(
                investigation, "__volume"
            ),
            sample_count=self.__get_investigation_parameter_by_name(
                investigation, "__sampleCount"
            ),
            dataset_count=self.__get_investigation_parameter_by_name(
                investigation, "__datasetCount"
            ),
        )

    def get_full_user_name(self):
        return self.icat_session["fullName"]

    def get_user_name(self):
        return self.icat_session["username"]

    def to_sessions(self, investigations):
        return [self.__to_session(investigation) for investigation in investigations]

    def get_parcels(self):
        """Returns the parcels associated to an investigation"""
        try:
            session_id = self.session_manager.active_session.session_id
            msg = f"Retrieving parcels by investigation_id {session_id}"
            logger.debug(msg)
            parcels = self.icatClient.get_parcels_by(session_id)
        except Exception:
            logger.exception("Failed on get_parcels_by_investigation_id")
        else:
            msg = f"Successfully retrieved {len(parcels)} parcels"
            logger.debug(msg)
            return parcels

        return []

    def get_samples_sheets(self) -> List[SampleSheet]:
        """Returns the samples sheets associated to an investigation"""
        try:
            msg = "Retrieving samples by investigation_id "
            msg += f"{self.session_manager.active_session.session_id}"
            logger.debug(msg)
            samples = self.icatClient.get_samples_by(
                self.session_manager.active_session.session_id
            )
        except Exception:
            logger.exception("Failed on get_samples_by_investigation_id")
        else:
            msg = f"Successfully retrieved {len(samples)} samples"
            logger.debug(msg)
            # Convert to object
            return [SampleSheet.parse_obj(sample) for sample in samples]
        return []

    def echo(self):
        """Mockup for the echo method."""
        return True

    def is_connected(self):
        return self.login_ok

    def add_beamline_configuration_metadata(self, metadata, beamline_config):
        """
        This is the mapping betweeh the beamline_config dict and the ICAt keys
        in case they exist then they will be added to the metadata of the dataset
        """
        if beamline_config is not None:
            key_mapping = {
                "detector_px": "InstrumentDetector01_beam_center_x",
                "detector_py": "InstrumentDetector01_beam_center_y",
                "beam_divergence_vertical": (
                    "InstrumentBeam_vertical_incident_beam_divergence"
                ),
                "beam_divergence_horizontal": (
                    "InstrumentBeam_horizontal_incident_beam_divergence"
                ),
                "polarisation": "InstrumentBeam_final_polarization",
                "detector_model": "InstrumentDetector01_model",
                "detector_manufacturer": "InstrumentDetector01_manufacturer",
                "synchrotron_name": "InstrumentSource_name",
                "monochromator_type": "InstrumentMonochromatorCrystal_type",
                "InstrumentDetector01_type": "detector_type",
            }

            for config_key, metadata_key in key_mapping.items():
                if hasattr(beamline_config, config_key):
                    metadata[metadata_key] = getattr(beamline_config, config_key)

    def find_sample_by_sample_id(self, sample_id):
        return next(
            (
                sample
                for sample in self.samples
                if str(sample["limsID"]) == str(sample_id)
            ),
            None,
        )

    def _get_sample_position(self) -> tuple:
        """Return the position of the puck in the samples changer
        and the position of the sample within the puck,
        """
        try:
            queue_entry = HWR.beamline.queue_manager.get_current_entry()
            sample_node = queue_entry.get_data_model().get_sample_node()
            location = sample_node.location  # Example: (8,2,5)

            if len(location) == 3:
                (cell, puck, sample_position) = location
            else:
                cell = 1
                (puck, sample_position) = location

            position = None
            if None not in (cell, puck):
                position = int(cell * 3) + int(puck)
        except Exception:
            logger.exception("Cannot retrieve sample position")
        return position, sample_position

    def store_beamline_setup(self, session_id: str, bl_config_dict: dict):
        pass

    def store_image(self, image_dict: dict):
        pass

    def store_common_data(self, datacollection_dict: dict) -> dict:
        """Fill in a dictionary with the common for all the
           data collection tecjniques meta data.
        Args:
            datacollection_dict(dict): dictionarry from the data collection.
        """
        sample_id = datacollection_dict.get("blSampleId")
        msg = f"SampleId is: {sample_id}"
        self.log.debug(msg)
        try:
            sample = HWR.beamline.lims.find_sample_by_sample_id(sample_id)
            sample_name = sample.get("sampleName")
        except (AttributeError, TypeError):
            sample_name = "unknown"
            msg = f"Sample {sample_id} not found"
            self.log.debug(msg)

        start_time = datacollection_dict.get("collection_start_time", "")
        end_time = datetime.now(ZoneInfo("Europe/Paris")).isoformat()

        if start_time:
            try:
                dt_aware = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S").replace(
                    tzinfo=ZoneInfo("Europe/Paris")
                )
                start_time = dt_aware.isoformat(timespec="microseconds")
            except (ValueError, TypeError):
                self.log.exception("Cannot parse start time")
        else:
            start_time = datetime.now(ZoneInfo("Europe/Paris")).isoformat()

        bsx, bsy, shape, _ = HWR.beamline.beam.get_value()
        flux_end = datacollection_dict.get("flux_end") or HWR.beamline.flux.get_value()

        transmission = (
            datacollection_dict.get("transmission")
            or HWR.beamline.transmission.get_value()
        )

        energy = datacollection_dict.get("energy") or HWR.beamline.energy.get_value()
        wavelength = (
            datacollection_dict.get("wavelength")
            or HWR.beamline.energy.get_wavelength()
        )

        machine_info = HWR.beamline.machine_info.get_value()

        # name of the beamline where the experiment is being conducted
        beamline_name = HWR.beamline.session.beamline_name
        # name of the beamline where the experiment was scheduled
        scheduled_beamline_name = HWR.beamline.session.beamline_name
        try:
            scheduled_beamline_name = self._get_scheduled_beamline()
            msg += f"Current Beamline={HWR.beamline.session.beamline_name}"
            logger.info(msg)
        except RuntimeError as err:
            msg = f"Failed to get _get_scheduled_beamline {err}"
            logger.warning(msg)

        investigation_id = None
        investigation_name = None
        if self.session_manager.active_session.session_id:
            investigation_id = self.session_manager.active_session.session_id
            session = self.get_session_by_id(investigation_id)
            if session is not None:
                investigation_name = session.proposal_name

        return {
            "sampleId": sample_id,
            "Sample_name": sample_name,
            "startDate": start_time,
            "endDate": end_time,
            "investigation_id": investigation_id,
            "beamline_name": beamline_name,
            "proposal": investigation_name,
            "scheduled_beamline_name": scheduled_beamline_name,
            "MX_beamShape": shape.value,
            "MX_beamSizeAtSampleX": bsx,
            "MX_beamSizeAtSampleY": bsy,
            "MX_xBeam": datacollection_dict.get("xBeam", "N/A"),
            "MX_yBeam": datacollection_dict.get("yBeam", "N/A"),
            "MX_flux": datacollection_dict.get("flux"),
            "MX_fluxEnd": flux_end,
            "MX_transmission": transmission,
            "InstrumentMonochromator_wavelength": wavelength,
            "InstrumentMonochromator_energy": energy,
            "InstrumentSource_current": machine_info.get("current"),
            "InstrumentSource_mode": machine_info.get("fill_mode"),
        }

    def store_energy_scan(self, energyscan_dict: dict):
        try:
            metadata = self.store_common_data(energyscan_dict)
            try:
                beamline = self._get_scheduled_beamline()
                msg = f"Dataset Beamline={beamline} "
                msg += f"Current Beamline={HWR.beamline.session.beamline_name}"
                self.log.info(msg)
            except Exception:
                self.log.exception(
                    "Failed to get _get_scheduled_beamline",
                )
            _session = HWR.beamline.session
            proposal = f"{_session.proposal_code}{_session.proposal_number}"

            directory = Path(energyscan_dict["scanFileFullPath"]).parent

            start_time = energyscan_dict.get("startTime", "")
            end_time = energyscan_dict.get("endTime", "")

            if start_time:
                try:
                    dt_aware = datetime.strptime(
                        start_time, "%Y-%m-%d %H:%M:%S"
                    ).replace(tzinfo=ZoneInfo("Europe/Paris"))
                    start_time = dt_aware.isoformat(timespec="microseconds")
                    metadata.update({"startDate": start_time})
                except (ValueError, TypeError):
                    self.log.exception("Cannot parse start time")

            if end_time:
                try:
                    dt_aware = datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S").replace(
                        tzinfo=ZoneInfo("Europe/Paris")
                    )
                    end_time = dt_aware.isoformat(timespec="microseconds")
                    metadata.update({"endDate": end_time})
                except (ValueError, TypeError):
                    self.log.exception("Cannot parse start time")

            metadata.update(
                {
                    "scanType": "energy_scan",
                    "MX_directory": str(directory),
                    "MX_exposureTime": energyscan_dict.get("exposureTime"),
                    "InstrumentDetector01_model": energyscan_dict.get(
                        "fluorescenceDetector"
                    ),
                    "MX_element": energyscan_dict.get("element"),
                    "MX_edgeEnergy": energyscan_dict.get("edgeEnergy"),
                    "MX_startEnergy": energyscan_dict.get("startEnergy"),
                    "MX_endEnergy": energyscan_dict.get("endEnergy"),
                    "MX_peakEnergy": energyscan_dict.get("endEnergy"),
                    "MX_inflectioEnergy": energyscan_dict.get("inflectioEnergy"),
                    "MX_remoteEnergy": energyscan_dict.get("remoteEnergy"),
                    "MX_peakFPrime": energyscan_dict.get("peakFPrime"),
                    "MX_peakFDoublePrime": energyscan_dict.get("peakFDoublePrime"),
                    "MX_inflectionFPrime": energyscan_dict.get("inflectionFPrime"),
                    "MX_inflectionFDoublePrime": energyscan_dict.get(
                        "inflectionFDoublePrime"
                    ),
                    "MX_comments": energyscan_dict.get("comments"),
                }
            )

            self.icatClient.store_dataset(
                beamline=beamline,
                proposal=proposal,
                dataset=str(directory.name),
                path=str(directory),
                metadata=metadata,
            )
        except Exception:
            logging.getLogger("ispyb_client").exception()

    def store_xfe_spectrum(self, xfespectrum_dict: dict):
        status = {"xfeFluorescenceSpectrumId": -1}
        try:
            metadata = self.store_common_data(xfespectrum_dict)
            try:
                beamline = self._get_scheduled_beamline()
                msg = f"Dataset Beamline={beamline} "
                msg += f"Current Beamline={HWR.beamline.session.beamline_name}"
                self.log.info(msg)
            except Exception:
                self.log.exception(
                    "Failed to get _get_scheduled_beamline",
                )
            _session = HWR.beamline.session
            proposal = f"{_session.proposal_code}{_session.proposal_number}"

            directory = Path(xfespectrum_dict["filename"]).parent

            start_time = xfespectrum_dict.get("startTime", "")
            end_time = xfespectrum_dict.get("endTime", "")

            if start_time:
                try:
                    dt_aware = datetime.strptime(
                        start_time, "%Y-%m-%d %H:%M:%S"
                    ).replace(tzinfo=ZoneInfo("Europe/Paris"))
                    start_time = dt_aware.isoformat(timespec="microseconds")
                    metadata.update({"startDate": start_time})
                except (ValueError, TypeError):
                    self.log.exception("Cannot parse start time")

            if end_time:
                try:
                    dt_aware = datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S").replace(
                        tzinfo=ZoneInfo("Europe/Paris")
                    )
                    end_time = dt_aware.isoformat(timespec="microseconds")
                    metadata.update({"endDate": end_time})
                except (ValueError, TypeError):
                    self.log.exception("Cannot parse end time")

            metadata.update(
                {
                    "scanType": "xrf",
                    "MX_directory": str(directory),
                    "MX_exposureTime": xfespectrum_dict.get("exposureTime"),
                }
            )

            self.icatClient.store_dataset(
                beamline=beamline,
                proposal=proposal,
                dataset=str(directory.name),
                path=str(directory),
                metadata=metadata,
            )
        except Exception:
            logging.getLogger("ispyb_client").exception()

        return status

    def store_workflow(self, workflow_dict: dict):
        pass

    def store_data_collection(self, datacollection_dict, beamline_config_dict=None):
        """Store the dictionary with the information about the beamline
        to be sent when a dataset is produced.
        """
        self.beamline_config = beamline_config_dict

    def update_data_collection(self, datacollection_dict: dict):
        """Update data collection."""

    def _get_oscillation_end(self, oscillation_sequence):
        return float(oscillation_sequence["start"]) + (
            float(oscillation_sequence["range"])
            - float(oscillation_sequence["overlap"])
        ) * float(oscillation_sequence["number_of_images"])

    def _get_rotation_axis(self, oscillation_sequence):
        if "kappaStart" in oscillation_sequence:
            if (
                oscillation_sequence["kappaStart"] != 0
                and oscillation_sequence["kappaStart"] != -9999
            ):
                return "Omega"
        return "Phi"

    def __get_sample_information_by(
        self, sample_id: str
    ) -> Optional[SampleInformation]:
        """
        Fetches sample metadata and associated resources based on the sample ID.

        Parameters:
            sample_id (str): The unique identifier for the sample.

        Returns:
            Optional[SampleInformation]: Returns a SampleInformation object or None.
        """
        try:
            result = self.icatClient.get_sample_files_information_by(sample_id)
            return SampleInformation.parse_obj(result)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logger.info("Sample %s not found (404)", sample_id)
            else:
                logger.exception("HTTP error for sample %s", sample_id)
        except requests.exceptions.RequestException:
            logger.exception("Request error for sample %s", sample_id)
        except ValidationError:
            logger.exception("Invalid response format for sample %s", sample_id)
        return None

    def _download_resources(
        self, sample_id, resources, output_folder: str, sample_name: str
    ) -> List[Download]:
        """
        Download resources related to a given sample and save them to the
        specified directory.

        Argss:
            sample_id (str): Sample identifier.
            output_folder (str): Directory where storefiles will be saved.

        Returns:
            List containing the paths of the downloaded files.
        """
        downloaded_files: List[Download] = []
        for resource in resources:
            resource_folder = Path(output_folder) / sample_name
            resource_folder = Path(resource_folder) / (
                resource.groupName if resource.groupName else ""
            )
            resource_folder.mkdir(
                parents=True,
                exist_ok=True,
            )  # Make sure the folder exists

            try:
                result = self.icatClient.download_file_by(
                    sample_id=sample_id,
                    resource_id=resource.id,
                    use_chunks=True,
                    chunk_size=8192,
                )
                output_path = Path(resource_folder / resource.filename)
                with output_path.open("wb") as f:
                    f.write(result)

                # Create a new Download instance with updated path
                downloaded = Download(
                    path=str(output_path),
                    filename=resource.filename,
                    groupName=resource.groupName,
                )
                downloaded_files.append(downloaded)
                logger.info("Downloaded %s to %s", resource.filename, downloaded.path)

            except requests.exceptions.RequestException:
                logger.exception("Failed to download %s", resource.filename)

        return downloaded_files

    def finalize_data_collection(self, datacollection_dict):
        logger.info("Storing datacollection in ICAT")

        metadata = self.store_common_data(datacollection_dict)

        try:
            fileinfo = datacollection_dict["fileinfo"]
            directory = Path(fileinfo["directory"])
            dataset_name = directory.name
            # Determine the scan type
            scan_types = ["mesh", "line", "characterisation", "datacollection"]
            scan_type = datacollection_dict["experiment_type"]
            for nam in scan_types:
                if dataset_name.endswith(nam):
                    scan_type = nam

            workflow_params = datacollection_dict.get("workflow_parameters", {})
            workflow_type = workflow_params.get("workflow_type")

            if workflow_type is None and not directory.name.startswith("run"):
                dataset_name = fileinfo["prefix"]

            if datacollection_dict["sample_reference"]["acronym"]:
                sample_name = (
                    datacollection_dict["sample_reference"]["acronym"]
                    + "-"
                    + datacollection_dict["sample_reference"]["sample_name"]
                )
            else:
                sample_name = datacollection_dict["sample_reference"][
                    "sample_name"
                ].replace(":", "-")

            logger.info(f"LIMS sample name {sample_name}")
            oscillation_sequence = datacollection_dict["oscillation_sequence"][0]

            beamline = HWR.beamline.session.beamline_name.lower()
            distance = HWR.beamline.detector.distance.get_value()
            proposal = f"{HWR.beamline.session.proposal_code}"
            proposal += f"{HWR.beamline.session.proposal_number}"

            metadata.update(
                {
                    "MX_dataCollectionId": datacollection_dict.get("collection_id"),
                    "MX_detectorDistance": distance,
                    "MX_directory": str(directory),
                    "MX_exposureTime": oscillation_sequence["exposure_time"],
                    "MX_positionName": datacollection_dict.get("position_name"),
                    "MX_numberOfImages": oscillation_sequence["number_of_images"],
                    "MX_oscillationRange": oscillation_sequence["range"],
                    "MX_axis_start": oscillation_sequence["start"],
                    "MX_oscillationOverlap": oscillation_sequence["overlap"],
                    "MX_resolution": datacollection_dict.get("resolution"),
                    "MX_resolution_at_corner": datacollection_dict.get(
                        "resolutionAtCorner"
                    ),
                    "scanType": scan_type,
                    "MX_startImageNumber": oscillation_sequence["start_image_number"],
                    "MX_template": fileinfo["template"],
                    "Sample_name": sample_name,
                    "Workflow_name": workflow_params.get("workflow_name"),
                    "Workflow_type": workflow_params.get("workflow_type"),
                    "Workflow_id": workflow_params.get("workflow_uid"),
                    "MX_kappa_settings_id": workflow_params.get(
                        "workflow_kappa_settings_id"
                    ),
                    "MX_characterisation_id": workflow_params.get(
                        "workflow_characterisation_id"
                    ),
                    "MX_position_id": workflow_params.get("workflow_position_id"),
                    "group_by": workflow_params.get("workflow_group_by"),
                }
            )

            metadata["SampleTrackingContainer_type"] = "UNIPUCK"
            metadata["SampleTrackingContainer_capacity"] = "16"
            (position, sample_position) = self._get_sample_position()
            metadata["SampleChanger_position"] = position
            metadata["SampleTrackingContainer_position"] = sample_position
            # Find sample by sampleId
            sample = HWR.beamline.lims.find_sample_by_sample_id(
                datacollection_dict.get("blSampleId")
            )

            try:
                metadata["lims"] = HWR.beamline.lims.get_active_lims().name
            except Exception:
                logger.exception("Failed to read get_active_lims.")

            try:
                if sample is not None:
                    metadata["SampleProtein_acronym"] = sample.get("proteinAcronym")
                    metadata["SampleTrackingContainer_id"] = sample.get(
                        "containerCode"
                    )  # containerCode instead of sampletrackingcontainer_id
                    # for ISPyB's compatiblity
                    metadata["SampleTrackingParcel_id"] = sample.get(
                        "SampleTrackingParcel_id"
                    )
                    metadata["SampleTrackingParcel_name"] = sample.get(
                        "SampleTrackingParcel_name"
                    )
            except RuntimeError as e:
                logger.warning("Failed to add sample metadata.%s", e)

            try:
                self.add_beamline_configuration_metadata(metadata, self.beamline_config)
            except RuntimeError as e:
                logger.warning("Failed to add_beamline_configuration_metadata.%s", e)

            # MX_axis_end
            try:
                metadata["MX_axis_end"] = self._get_oscillation_end(
                    oscillation_sequence
                )
            except RuntimeError:
                logger.warning("Failed to get MX_axis_end")

            # MX_axis_end
            try:
                metadata["MX_axis_range"] = self._get_rotation_axis(
                    oscillation_sequence
                )
            except RuntimeError:
                logger.warning("Failed to get MX_axis_end")

            icat_metadata_path = Path(directory) / "metadata.json"
            with Path(icat_metadata_path).open("w") as f:
                # We add the processing and experiment plan only in the metadata.json
                # it will not work thought pyicat-plus
                merged = metadata.copy()
                try:
                    if sample is not None:
                        merged["experimentPlan"] = sample.get("experimentPlan")
                        merged["processingPlan"] = sample.get("processingPlan")
                except RuntimeError as e:
                    logger.warning("Failed to get merged sample plan. %s", e)

                f.write(json.dumps(merged, indent=4))

            # Create ICAT gallery
            try:
                gallery_path = directory / "gallery"
                gallery_path.mkdir(mode=0o755, exist_ok=True)
                for snapshot_index in range(1, 5):
                    key = f"xtalSnapshotFullPath{snapshot_index}"
                    if key in datacollection_dict:
                        snapshot_path = Path(datacollection_dict[key])
                        if snapshot_path.exists():
                            msg = f"Copying snapshot index {snapshot_index} to gallery"
                            logger.debug(msg)
                            shutil.copy(snapshot_path, gallery_path)
            except RuntimeError as e:
                logger.warning("Failed to create gallery. %s", e)

            self.icatClient.store_dataset(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset_name,
                path=str(directory),
                metadata=metadata,
            )
            logger.debug("Done uploading to ICAT")
        except Exception as e:
            logger.warning("Failed uploading to ICAT. %s", e)

    def _get_scheduled_beamline(self) -> str:
        """Return the name of the beamline as set in the properties or the
        name of the beamline where the session has been scheduled
        (in case of a different beamline)
        """
        active_session = self.session_manager.active_session

        if active_session is None or active_session.is_scheduled_beamline:
            return HWR.beamline.session.beamline_name.lower()

        beamline = str(active_session.beamline_name.lower())
        msg = f"Session have been moved to another beamline: {beamline}"
        logger.info(msg)
        return beamline

    def update_bl_sample(self, bl_sample: str):
        """
        Creates or stos a BLSample entry.
        # NBNB update doc string
        :param sample_dict: A dictionary with the properties for the entry.
        :type sample_dict: dict
        """
